import os

from audiomate import corpus
from audiomate import issuers
from audiomate.corpus.io import audio_mnist

from tests import resources
from . import reader_test as rt


class TestAudioMNISTReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('audio_mnist')
    FILE_TRACK_BASE_PATH = os.path.join(SAMPLE_PATH, 'data')

    EXPECTED_NUMBER_OF_TRACKS = 7
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('0_01_0', '01/0_01_0.wav'),
        rt.ExpFileTrack('0_01_1', '01/0_01_1.wav'),
        rt.ExpFileTrack('1_01_0', '01/1_01_0.wav'),
        rt.ExpFileTrack('0_02_0', '02/0_02_0.wav'),
        rt.ExpFileTrack('4_03_0', '03/4_03_0.wav'),
        rt.ExpFileTrack('4_03_1', '03/4_03_1.wav'),
        rt.ExpFileTrack('0_04_0', '04/0_04_0.wav'),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 4
    EXPECTED_ISSUERS = [
        rt.ExpSpeaker('01', 3, issuers.Gender.MALE,
                      issuers.AgeGroup.ADULT, None),
        rt.ExpSpeaker('02', 1, issuers.Gender.MALE,
                      issuers.AgeGroup.ADULT, None),
        rt.ExpSpeaker('03', 2, issuers.Gender.FEMALE,
                      issuers.AgeGroup.ADULT, None),
        rt.ExpSpeaker('04', 1, issuers.Gender.MALE,
                      issuers.AgeGroup.YOUTH, None),
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 7
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('0_01_0', '0_01_0', '01', 0, float('inf')),
        rt.ExpUtterance('0_01_1', '0_01_1', '01', 0, float('inf')),
        rt.ExpUtterance('1_01_0', '1_01_0', '01', 0, float('inf')),
        rt.ExpUtterance('0_02_0', '0_02_0', '02', 0, float('inf')),
        rt.ExpUtterance('4_03_0', '4_03_0', '03', 0, float('inf')),
        rt.ExpUtterance('4_03_1', '4_03_1', '03', 0, float('inf')),
        rt.ExpUtterance('0_04_0', '0_04_0', '04', 0, float('inf')),
    ]

    EXPECTED_LABEL_LISTS = {
        '0_01_0': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
        '1_01_0': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
        '4_03_0': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1),
        ],
        '0_04_0': [
            rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)
        ],
    }

    EXPECTED_LABELS = {
        '0_01_0': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, '0', 0, float('inf')),
        ],
        '1_01_0': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, '1', 0, float('inf')),
        ],
        '4_03_0': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, '4', 0, float('inf')),
        ],
        '0_04_0': [
            rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, '0', 0, float('inf')),
        ],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 0

    def load(self):
        reader = audio_mnist.AudioMNISTReader()
        return reader.load(self.SAMPLE_PATH)



import os

from audiomate import corpus
from audiomate import issuers
from audiomate.corpus import io
from audiomate.corpus.io import musan

import pytest
import requests_mock

from tests import resources
from . import reader_test as rt


@pytest.fixture()
def tar_data():
    with open(resources.get_resource_path(['sample_files', 'cv_corpus_v1.tar.gz']), 'rb') as f:
        return f.read()


class TestMusanDownloader:

    def test_download(self, tar_data, tmpdir):
        target_folder = tmpdir.strpath
        downloader = io.MusanDownloader()

        with requests_mock.Mocker() as mock:
            mock.get(musan.DOWNLOAD_URL, content=tar_data)

            downloader.download(target_folder)

        assert os.path.isfile(os.path.join(target_folder, 'cv-valid-dev.csv'))
        assert os.path.isdir(os.path.join(target_folder, 'cv-valid-dev'))
        assert os.path.isfile(os.path.join(target_folder, 'cv-valid-train.csv'))
        assert os.path.isdir(os.path.join(target_folder, 'cv-valid-train'))


class TestMusanReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('musan')
    FILE_TRACK_BASE_PATH = SAMPLE_PATH

    EXPECTED_NUMBER_OF_TRACKS = 5
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('music-fma-0000', os.path.join('music', 'fma', 'music-fma-0000.wav')),
        rt.ExpFileTrack('noise-free-sound-0000', os.path.join('noise', 'free-sound', 'noise-free-sound-0000.wav')),
        rt.ExpFileTrack('noise-free-sound-0001', os.path.join('noise', 'free-sound', 'noise-free-sound-0001.wav')),
        rt.ExpFileTrack('speech-librivox-0000', os.path.join('speech', 'librivox', 'speech-librivox-0000.wav')),
        rt.ExpFileTrack('speech-librivox-0001', os.path.join('speech', 'librivox', 'speech-librivox-0001.wav')),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 3
    EXPECTED_ISSUERS = [
        rt.ExpSpeaker('speech-librivox-0000', 1, issuers.Gender.MALE, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('speech-librivox-0001', 1, issuers.Gender.FEMALE, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpArtist('Quiet_Music_for_Tiny_Robots', 1, 'Quiet_Music_for_Tiny_Robots'),
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 5
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('music-fma-0000', 'music-fma-0000', 'Quiet_Music_for_Tiny_Robots', 0, float('inf')),
        rt.ExpUtterance('noise-free-sound-0000', 'noise-free-sound-0000', None, 0, float('inf')),
        rt.ExpUtterance('noise-free-sound-0001', 'noise-free-sound-0001', None, 0, float('inf')),
        rt.ExpUtterance('speech-librivox-0000', 'speech-librivox-0000', 'speech-librivox-0000', 0, float('inf')),
        rt.ExpUtterance('speech-librivox-0001', 'speech-librivox-0001', 'speech-librivox-0001', 0, float('inf')),
    ]

    EXPECTED_LABEL_LISTS = {
        'music-fma-0000': [
            rt.ExpLabelList(corpus.LL_DOMAIN, 1),
        ],
        'noise-free-sound-0000': [
            rt.ExpLabelList(corpus.LL_DOMAIN, 1),
        ],
        'noise-free-sound-0001': [
            rt.ExpLabelList(corpus.LL_DOMAIN, 1),
        ],
    }

    EXPECTED_LABELS = {
        'music-fma-0000': [
            rt.ExpLabel(corpus.LL_DOMAIN, 'music', 0, float('inf')),
        ],
        'noise-free-sound-0000': [
            rt.ExpLabel(corpus.LL_DOMAIN, 'noise', 0, float('inf')),
        ],
        'noise-free-sound-0001': [
            rt.ExpLabel(corpus.LL_DOMAIN, 'noise', 0, float('inf')),
        ],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 0

    def load(self):
        return io.MusanReader().load(self.SAMPLE_PATH)



import numpy as np

from audiomate.processing import pipeline


class TestAddContext:

    def test_compute_with_left_frames(self):
        data = np.array([[1, 2], [3, 4], [5, 6]])
        chunk = pipeline.Chunk(data, offset=0, is_last=True)

        step = pipeline.AddContext(left_frames=2, right_frames=0)
        result = step.compute(chunk, 16000)

        assert np.array_equal(result, np.array([
            [0, 0, 0, 0, 1, 2],
            [0, 0, 1, 2, 3, 4],
            [1, 2, 3, 4, 5, 6]
        ]))

    def test_compute_with_right_frames(self):
        data = np.array([[1, 2], [3, 4], [5, 6]])
        chunk = pipeline.Chunk(data, offset=0, is_last=True)

        step = pipeline.AddContext(left_frames=0, right_frames=2)
        result = step.compute(chunk, 16000)

        assert np.array_equal(result, np.array([
            [1, 2, 3, 4, 5, 6],
            [3, 4, 5, 6, 0, 0],
            [5, 6, 0, 0, 0, 0]
        ]))

    def test_compute_with_left_and_right_frames(self):
        data = np.array([[1, 2], [3, 4], [5, 6]])
        chunk = pipeline.Chunk(data, offset=0, is_last=True)

        step = pipeline.AddContext(left_frames=2, right_frames=2)
        result = step.compute(chunk, 16000)

        assert np.array_equal(result, np.array([
            [0, 0, 0, 0, 1, 2, 3, 4, 5, 6],
            [0, 0, 1, 2, 3, 4, 5, 6, 0, 0],
            [1, 2, 3, 4, 5, 6, 0, 0, 0, 0]
        ]))

    def test_compute_online_with_left_frames(self):
        step = pipeline.AddContext(left_frames=2, right_frames=0)

        # FIRST CHUNK
        data = np.array([[1, 2], [3, 4]])
        result = step.process_frames(data, 16000, offset=0, last=False)

        assert np.array_equal(result, np.array([
            [0, 0, 0, 0, 1, 2],
            [0, 0, 1, 2, 3, 4],
        ]))

        # SECOND CHUNK
        data = np.array([[5, 6], [7, 8]])
        result = step.process_frames(data, 16000, offset=2, last=True)

        assert np.array_equal(result, np.array([
            [1, 2, 3, 4, 5, 6],
            [3, 4, 5, 6, 7, 8]
        ]))

    def test_compute_online_with_right_frames(self):
        step = pipeline.AddContext(left_frames=0, right_frames=2)

        # FIRST CHUNK
        data = np.array([[1, 2], [3, 4]])
        result = step.process_frames(data, 16000, offset=0, last=False)

        # Since it waits on context, no output is expected
        assert result is None

        # SECOND CHUNK
        data = np.array([[5, 6], [7, 8]])
        result = step.process_frames(data, 16000, offset=2, last=True)

        assert np.array_equal(result, np.array([
            [1, 2, 3, 4, 5, 6],
            [3, 4, 5, 6, 7, 8],
            [5, 6, 7, 8, 0, 0],
            [7, 8, 0, 0, 0, 0]
        ]))

    def test_compute_online_with_left_and_right_frames(self):
        step = pipeline.AddContext(left_frames=2, right_frames=2)

        # FIRST CHUNK
        data = np.array([[1, 2], [3, 4]])
        result = step.process_frames(data, 16000, offset=0, last=False)

        # Since it waits on context, no output is expected
        assert result is None

        # SECOND CHUNK
        data = np.array([[5, 6], [7, 8]])
        result = step.process_frames(data, 16000, offset=2, last=True)

        assert np.array_equal(result, np.array([
            [0, 0, 0, 0, 1, 2, 3, 4, 5, 6],
            [0, 0, 1, 2, 3, 4, 5, 6, 7, 8],
            [1, 2, 3, 4, 5, 6, 7, 8, 0, 0],
            [3, 4, 5, 6, 7, 8, 0, 0, 0, 0]
        ]))



import os
import glob
import re

import audiomate
from audiomate import annotations
from audiomate.corpus import subset
from . import base
from . import downloader

DATA_URL = 'https://data.vision.ee.ethz.ch/cvl/ae_dataset/ae_dataset.zip'
LABEL_PATTERN = re.compile(r'(.*)_\d')


class AEDDownloader(downloader.ArchiveDownloader):
    """
    Downloader for the Acoustic Event Dataset.
    """

    def __init__(self):
        super(AEDDownloader, self).__init__(
            DATA_URL,
            move_files_up=True
        )

    @classmethod
    def type(cls):
        return 'aed'


class AEDReader(base.CorpusReader):
    """
    Reader for the Acoustic Event Dataset.

    .. seealso::

       `AED <https://data.vision.ee.ethz.ch/cvl/ae_dataset/>`_
          Download page
    """

    @classmethod
    def type(cls):
        return 'aed'

    def _check_for_missing_files(self, path):
        return []

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)

        test_folder = os.path.join(path, 'test')
        train_folder = os.path.join(path, 'train')

        test_utterance_ids = AEDReader.load_folder(test_folder, corpus)
        train_utterance_ids = AEDReader.load_folder(train_folder, corpus)

        test_filter = subset.MatchingUtteranceIdxFilter(utterance_idxs=test_utterance_ids)
        train_filter = subset.MatchingUtteranceIdxFilter(utterance_idxs=train_utterance_ids)

        test_subset = subset.Subview(corpus, filter_criteria=[test_filter])
        train_subset = subset.Subview(corpus, filter_criteria=[train_filter])

        corpus.import_subview('test', test_subset)
        corpus.import_subview('train', train_subset)

        return corpus

    @staticmethod
    def load_folder(path, corpus):
        utterance_ids = set()

        for wav_path in glob.glob(os.path.join(path, '**/*.wav'), recursive=True):
            basename = os.path.splitext(os.path.basename(wav_path))[0]

            match = LABEL_PATTERN.match(basename)

            if match is not None:
                label = match.group(1)

                corpus.new_file(wav_path, basename)
                utt = corpus.new_utterance(basename, basename)
                label = annotations.LabelList.create_single(label, audiomate.corpus.LL_SOUND_CLASS)
                utt.set_label_list(label)
                utterance_ids.add(basename)

        return utterance_ids



import os.path

import pytest

from audiomate.annotations import Label, LabelList
from audiomate.annotations import relabeling


class TestLabelListUtilities:

    def test_relabel_maps_a_onto_b(self):
        label_list = LabelList(labels=[
            Label('a', 3.2, 4.5)
        ])

        actual = relabeling.relabel(label_list, {('a',): 'b'})

        expected = LabelList(labels=[
            Label('b', 3.2, 4.5)
        ])

        assert actual == expected

    def test_relabel_flattens_partial_overlap_into_combined_label(self):
        projections = {
            ('a',): 'a',
            ('b',): 'b',
            ('c',): 'c',
            ('a', 'b',): 'a_b',
            ('a', 'b', 'c',): 'a_b_c',
            ('b', 'c',): 'b_c',
        }

        label_list = LabelList(labels=[
            Label('a', 3.2, 4.5),
            Label('b', 4.0, 4.9),
            Label('c', 4.2, 5.1)
        ])

        actual = relabeling.relabel(label_list, projections)

        expected = LabelList(labels=[
            Label('a', 3.2, 4.0),
            Label('a_b', 4.0, 4.2),
            Label('a_b_c', 4.2, 4.5),
            Label('b_c', 4.5, 4.9),
            Label('c', 4.9, 5.1)
        ])

        assert actual == expected

    def test_relabel_flattens_full_overlap_into_combined_label(self):
        projections = {
            ('a',): 'a',
            ('b',): 'b',
            ('a', 'b'): 'a_b',
        }

        label_list = LabelList(labels=[
            Label('a', 3.2, 4.9),
            Label('b', 3.9, 4.5)
        ])

        actual = relabeling.relabel(label_list, projections)

        expected = LabelList(labels=[
            Label('a', 3.2, 3.9),
            Label('a_b', 3.9, 4.5),
            Label('a', 4.5, 4.9)
        ])

        assert actual == expected

    def test_relabel_removes_unwanted_labels(self):
        projections = {
            ('a',): '',
            ('b',): 'b',
        }

        label_list = LabelList(labels=[
            Label('a', 3.2, 4.4),
            Label('b', 4.4, 5.1)
        ])

        actual = relabeling.relabel(label_list, projections)

        expected = LabelList(labels=[
            Label('b', 4.4, 5.1)
        ])

        assert actual == expected

    def test_relabel_removes_overlapping_segment(self):
        projections = {
            ('a',): 'a',
            ('a', 'b',): '',
            ('b',): 'b',
        }

        label_list = LabelList(labels=[
            Label('a', 3.2, 5.1),
            Label('b', 4.2, 4.7)
        ])

        actual = relabeling.relabel(label_list, projections)

        expected = LabelList(labels=[
            Label('a', 3.2, 4.2),
            Label('a', 4.7, 5.1)
        ])

        assert actual == expected

    def test_relabel_throws_error_if_unmapped_labels_are_detected(self):
        label_list = LabelList(labels=[
            Label('a', 3.2, 5.1),
            Label('b', 4.2, 4.7),
            Label('c', 4.3, 4.8)
        ])

        unmapped_combinations = [('a', 'b'), ('a', 'b', 'c'), ('a', 'c')]
        expected_message = 'Unmapped combinations: {}'.format(unmapped_combinations)

        with pytest.raises(relabeling.UnmappedLabelsException) as ex:
            relabeling.relabel(label_list, {('a',): 'foo'})

        assert ex.value.message == expected_message

    def test_relabel_proceeds_despite_unmapped_labels_in_presence_of_wildcard_rule(self):
        label_list = LabelList(labels=[
            Label('a', 3.2, 5.1),
            Label('b', 4.2, 4.7),
            Label('c', 4.3, 4.8)
        ])

        actual = relabeling.relabel(label_list, {('a',): 'new_label_a', ('**',): 'catch_all'})

        expected = LabelList(labels=[
            Label('new_label_a', 3.2, 4.2),
            Label('catch_all', 4.2, 4.3),
            Label('catch_all', 4.3, 4.7),
            Label('catch_all', 4.7, 4.8),
            Label('new_label_a', 4.8, 5.1),
        ])

        assert actual == expected

    def test_load_projections_from_file(self):
        path = os.path.join(os.path.dirname(__file__), 'projections.txt')
        projections = relabeling.load_projections(path)

        assert len(projections) == 3

        assert ('b',) in projections
        assert projections[('b',)] == 'foo'

        assert ('a', 'b',) in projections
        assert projections[('a', 'b',)] == 'a_b'

        assert ('a',) in projections
        assert projections[('a',)] == 'bar'

    def test_all_projections_missing_if_no_projections_defined(self):
        label_list = LabelList(labels=[
            Label('b', 3.2, 4.5),
            Label('a', 4.0, 4.9),
            Label('c', 4.2, 5.1)
        ])

        unmapped_combinations = relabeling.find_missing_projections(label_list, {})

        assert len(unmapped_combinations) == 5
        assert ('b',) in unmapped_combinations
        assert ('a', 'b',) in unmapped_combinations
        assert ('a', 'b', 'c',) in unmapped_combinations
        assert ('a', 'c',) in unmapped_combinations
        assert ('c',) in unmapped_combinations

    def test_all_missing_projections_found(self):
        projections = {
            ('a', 'b',): 'foo',
            ('c',): 'bar'
        }

        label_list = LabelList(labels=[
            Label('b', 3.2, 4.5),
            Label('a', 4.0, 4.9),
            Label('c', 4.2, 5.1)
        ])

        unmapped_combinations = relabeling.find_missing_projections(label_list, projections)

        assert len(unmapped_combinations) == 3
        assert ('b',) in unmapped_combinations
        assert ('a', 'b', 'c',) in unmapped_combinations
        assert ('a', 'c',) in unmapped_combinations

    def test_no_duplicate_missing_projections_reported(self):
        label_list = LabelList(labels=[
            Label('b', 1.0, 2.0),
            Label('a', 1.5, 2.5),
            Label('b', 3.0, 4.0),
            Label('a', 3.5, 4.5),
        ])

        unmapped_combinations = relabeling.find_missing_projections(label_list, {})

        assert len(unmapped_combinations) == 3
        assert ('b',) in unmapped_combinations
        assert ('a', 'b',) in unmapped_combinations
        assert ('a',) in unmapped_combinations

    def test_no_missing_projections_if_projection_complete(self):
        projections = {
            ('b',): 'foo',
            ('a', 'b',): 'foo',
            ('a', 'b', 'c',): 'foo',
            ('a', 'c',): 'foo',
            ('c',): 'bar'
        }

        label_list = LabelList(labels=[
            Label('b', 3.2, 4.5),
            Label('a', 4.0, 4.9),
            Label('c', 4.2, 5.1)
        ])

        unmapped_combinations = relabeling.find_missing_projections(label_list, projections)

        assert len(unmapped_combinations) == 0

    def test_no_missing_projections_if_covered_by_catch_all_rule(self):
        projections = {
            ('b',): 'new_label_b',
            ('**',): 'new_label_all',
        }

        label_list = LabelList(labels=[
            Label('b', 3.2, 4.5),
            Label('a', 4.0, 4.9),
            Label('c', 4.2, 5.1)
        ])

        unmapped_combinations = relabeling.find_missing_projections(label_list, projections)

        assert len(unmapped_combinations) == 0

    def test_missing_projections_are_naturally_sorted(self):
        label_list = LabelList(labels=[
            Label('b', 1.0, 2.0),
            Label('a', 1.5, 2.5),
        ])

        unmapped_combinations = relabeling.find_missing_projections(label_list, {})

        assert len(unmapped_combinations) == 3
        assert unmapped_combinations[0] == ('a',)
        assert unmapped_combinations[1] == ('a', 'b',)
        assert unmapped_combinations[2] == ('b',)



import os

from audiomate import issuers
from audiomate.corpus import io

from tests import resources
from . import reader_test as rt


class TestBroadcastReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('broadcast')
    FILE_TRACK_BASE_PATH = os.path.join(SAMPLE_PATH, 'files')

    EXPECTED_NUMBER_OF_TRACKS = 4
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('file-1', os.path.join('a', 'wav_1.wav')),
        rt.ExpFileTrack('file-2', os.path.join('b', 'wav_2.wav')),
        rt.ExpFileTrack('file-3', os.path.join('c', 'wav_3.wav')),
        rt.ExpFileTrack('file-4', os.path.join('d', 'wav_4.wav')),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 3
    EXPECTED_ISSUERS = [
        rt.ExpSpeaker('speaker-1', 2, issuers.Gender.FEMALE, issuers.AgeGroup.CHILD, 'eng'),
        rt.ExpArtist('speaker-2', 2, None),
        rt.ExpIssuer('speaker-3', 1),
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 5
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('utt-1', 'file-1', 'speaker-1', 0, float('inf')),
        rt.ExpUtterance('utt-2', 'file-2', 'speaker-1', 0, float('inf')),
        rt.ExpUtterance('utt-3', 'file-3', 'speaker-2', 0, 100),
        rt.ExpUtterance('utt-4', 'file-3', 'speaker-2', 100, 150),
        rt.ExpUtterance('utt-5', 'file-4', 'speaker-3', 0, float('inf')),
    ]

    EXPECTED_LABEL_LISTS = {
        'utt-1': [
            rt.ExpLabelList('jingles', 2),
            rt.ExpLabelList('music', 2),
        ],
        'utt-2': [
            rt.ExpLabelList('default', 2),
        ],
        'utt-3': [
            rt.ExpLabelList('default', 2),
        ],
        'utt-4': [
            rt.ExpLabelList('default', 2),
        ],
        'utt-5': [
            rt.ExpLabelList('default', 2),
        ],
    }

    EXPECTED_LABELS = {
        'utt-1': [
            rt.ExpLabel('jingles', 'velo', 80, 82.4),
        ],
        'utt-4': [
            rt.ExpLabel('default', 'hallo', 105, 130.5),
        ],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 0

    def test_load_label_meta(self):
        ds = self.load()

        utt_1 = ds.utterances['utt-1']
        labels = sorted(utt_1.label_lists['jingles'])

        assert len(labels[0].meta) == 0
        assert len(labels[1].meta) == 3
        assert labels[1].meta['lang'] == 'de'
        assert labels[1].meta['prio'] == 4
        assert labels[1].meta['unique']

    def load(self):
        return io.BroadcastReader().load(self.SAMPLE_PATH)



import librosa
import numpy as np

from . import base


class Delta(base.Computation):
    """
    Compute delta features.

    See http://librosa.github.io/librosa/generated/librosa.feature.delta.html
    """

    def __init__(self, width=9, order=1, axis=0, mode='interp', parent=None, name=None):
        needed_context = int(width / 2.0)

        super(Delta, self).__init__(parent=parent, name=name,
                                    min_frames=needed_context + 1,
                                    left_context=needed_context, right_context=needed_context)

        self.width = width
        self.order = order
        self.axis = axis
        self.mode = mode

    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        axis = len(chunk.data.shape) - self.axis - 1
        output = librosa.feature.delta(chunk.data.T, width=self.width, order=self.order, axis=axis, mode=self.mode).T

        return output[chunk.left_context:chunk.data.shape[0] - chunk.right_context]


class AddContext(base.Computation):
    """
    For every frame add context frames from left or/and right.
    For frames at the beginning and end of a sequence, where no context is available, zeros are used.

    Args:
        left_frames (int): Number of previous frames to prepend to a frame.
        right_frames (int): Number of subsequent frames to append to a frame.

    Example:
        >>> input = np.array([
        >>>     [1,2,3],
        >>>     [4,5,6],
        >>>     [7,8,9]
        >>> ])
        >>> chunk = Chunk(input, offset=0, is_last=True)
        >>> AddContext(left_frames=1, right_frames=1).compute(chunk, 16000)
        array([[0, 0, 0, 1, 2, 3, 4, 5, 6],
               [1, 2, 3, 4, 5, 6, 7, 8, 9],
               [4, 5, 6, 7, 8, 9, 0, 0, 0]])
    """

    def __init__(self, left_frames, right_frames, parent=None, name=None):
        super(AddContext, self).__init__(parent=parent, name=name, min_frames=1,
                                         left_context=left_frames, right_context=right_frames)

        self.left_frames = left_frames
        self.right_frames = right_frames

    def compute(self, chunk, sampling_rate, corpus=None, utterance=None):
        context = []

        for shift in range(self.left_frames, 0, -1):
            shift_context = chunk.data[:chunk.data.shape[0] - shift]
            pad_widths = [[0, 0] for __ in range(len(shift_context.shape))]
            pad_widths[0][0] = shift
            shift_context = np.pad(shift_context, pad_widths, mode='constant', constant_values=0)
            context.append(shift_context)

        context.append(chunk.data)

        for i in range(self.right_frames):
            shift = i + 1
            shift_context = chunk.data[shift:]
            pad_widths = [[0, 0] for __ in range(len(shift_context.shape))]
            pad_widths[0][1] = shift
            shift_context = np.pad(shift_context, pad_widths, mode='constant', constant_values=0)
            context.append(shift_context)

        stacked = np.hstack(context)
        return stacked[chunk.left_context:chunk.data.shape[0] - chunk.right_context]



from audiomate import annotations
from audiomate.utils import textfile

WILDCARD_COMBINATION = ('**',)


class UnmappedLabelsException(Exception):
    def __init__(self, message):
        super(Exception, self).__init__(message)
        self.message = message


def relabel(label_list, projections):
    """
    Relabel an entire :py:class:`~audiomate.annotations.LabelList` using user-defined projections.
    Labels can be renamed, removed or overlapping labels can be flattened to a single label per segment.

    Each entry in the dictionary of projections represents a single projection that maps a combination of labels (key)
    to a single new label (value). The combination of labels to be mapped is a tuple of naturally sorted labels that
    apply to one or more segments simultaneously. By defining a special wildcard projection using `('**',)` is is not
    required to specify a projection for every single combination of labels.

    This method raises a :py:class:`~audiomate.corpus.utils.labellist.UnmappedLabelsException` if a projection for one
    or more combinations of labels is not defined.

    Args:
        label_list (audiomate.annotations.LabelList): The label list to relabel
        projections (dict): A dictionary that maps tuples of label combinations to string
                            labels.
    Returns:
        audiomate.annotations.LabelList: New label list with remapped labels

    Raises:
        UnmappedLabelsException: If a projection for one or more combinations of labels is not defined.

    Example:
        >>> projections = {
        ...     ('a',): 'a',
        ...     ('b',): 'b',
        ...     ('c',): 'c',
        ...     ('a', 'b',): 'a_b',
        ...     ('a', 'b', 'c',): 'a_b_c',
        ...     ('**',): 'b_c',
        ... }
        >>> label_list = annotations.LabelList(labels=[
        ...     annotations.Label('a', 3.2, 4.5),
        ...     annotations.Label('b', 4.0, 4.9),
        ...     annotations.Label('c', 4.2, 5.1)
        ... ])
        >>> ll = relabel(label_list, projections)
        >>> [l.value for l in ll]
        ['a', 'a_b', 'a_b_c', 'b_c', 'c']
    """
    unmapped_combinations = find_missing_projections(label_list, projections)
    if len(unmapped_combinations) > 0:
        raise UnmappedLabelsException('Unmapped combinations: {}'.format(unmapped_combinations))

    new_labels = []
    for labeled_segment in label_list.ranges():
        combination = tuple(sorted([label.value for label in labeled_segment[2]]))
        label_mapping = projections[combination] if combination in projections else projections[WILDCARD_COMBINATION]

        if label_mapping == '':
            continue

        new_labels.append(annotations.Label(label_mapping, labeled_segment[0], labeled_segment[1]))

    return annotations.LabelList(idx=label_list.idx, labels=new_labels)


def find_missing_projections(label_list, projections):
    """
    Finds all combinations of labels in `label_list` that are not covered by an entry in the dictionary of
    `projections`. Returns a list containing tuples of uncovered label combinations or en empty list if there are none.
    All uncovered label combinations are naturally sorted.

    Each entry in the dictionary of projections represents a single projection that maps a combination of labels (key)
    to a single new label (value). The combination of labels to be mapped is a tuple of naturally sorted labels that
    apply to one or more segments simultaneously. By defining a special wildcard projection using `('**',)` is is not
    required to specify a projection for every single combination of labels.

    Args:
        label_list (audiomate.annotations.LabelList): The label list to relabel
        projections (dict): A dictionary that maps tuples of label combinations to string
                            labels.

    Returns:
        List: List of combinations of labels that are not covered by any projection

    Example:
        >>> ll = annotations.LabelList(labels=[
        ...     annotations.Label('b', 3.2, 4.5),
        ...     annotations.Label('a', 4.0, 4.9),
        ...     annotations.Label('c', 4.2, 5.1)
        ... ])
        >>> find_missing_projections(ll, {('b',): 'new_label'})
        [('a', 'b'), ('a', 'b', 'c'), ('a', 'c'), ('c',)]
    """
    unmapped_combinations = set()

    if WILDCARD_COMBINATION in projections:
        return []

    for labeled_segment in label_list.ranges():
        combination = tuple(sorted([label.value for label in labeled_segment[2]]))

        if combination not in projections:
            unmapped_combinations.add(combination)

    return sorted(unmapped_combinations)


def load_projections(projections_file):
    """
    Loads projections defined in the given `projections_file`.

    The `projections_file` is expected to be in the following format::

        old_label_1 | new_label_1
        old_label_1 old_label_2 | new_label_2
        old_label_3 |

    You can define one projection per line. Each projection starts with a list of one or multiple
    old labels (separated by a single whitespace) that are separated from the new label by a pipe
    (`|`). In the code above, the segment labeled with `old_label_1` will be labeled with
    `new_label_1` after applying the projection. Segments that are labeled with `old_label_1`
    **and** `old_label_2` concurrently are relabeled to `new_label_2`. All segments labeled with
    `old_label_3` are dropped. Combinations of multiple labels are automatically sorted in natural
    order.

    Args:
        projections_file (str): Path to the file with projections

    Returns:
        dict: Dictionary where the keys are tuples of labels to project to the key's value

    Example:
        >>> load_projections('/path/to/projections.txt')
        {('b',): 'foo', ('a', 'b'): 'a_b', ('a',): 'bar'}
    """
    projections = {}
    for parts in textfile.read_separated_lines_generator(projections_file, '|'):
        combination = tuple(sorted([label.strip() for label in parts[0].split(' ')]))
        new_label = parts[1].strip()

        projections[combination] = new_label

    return projections



import os
import tempfile

import audiomate
from audiomate.annotations import Label, LabelList
from audiomate.issuers import Issuer, Speaker, Gender
from audiomate.corpus.subset import subview


def get_resource_path(sub_path_components):
    """ Get the absolute path of a file in the resources folder with its relative path components. """
    return os.path.abspath(os.path.join(os.path.dirname(__file__), *sub_path_components))


def sample_wav_file(name):
    """ Return the path to a wav file of the `wav_files` folder with its name. """
    return get_resource_path(['wav_files', name])


def sample_corpus_path(name):
    """ Return the path to a sample corpus path with its name. """
    return get_resource_path(['sample_corpora', name])


def create_dataset():
    temp_path = tempfile.mkdtemp()

    ds = audiomate.Corpus(temp_path)

    wav_1_path = sample_wav_file('wav_1.wav')
    wav_2_path = sample_wav_file('wav_2.wav')
    wav_3_path = sample_wav_file('wav_3.wav')
    wav_4_path = sample_wav_file('wav_4.wav')

    file_1 = ds.new_file(wav_1_path, track_idx='wav-1')
    file_2 = ds.new_file(wav_2_path, track_idx='wav_2')
    file_3 = ds.new_file(wav_3_path, track_idx='wav_3')
    file_4 = ds.new_file(wav_4_path, track_idx='wav_4')

    issuer_1 = Speaker('spk-1', gender=Gender.MALE)
    issuer_2 = Speaker('spk-2', gender=Gender.FEMALE)
    issuer_3 = Issuer('spk-3')

    ds.import_issuers([issuer_1, issuer_2, issuer_3])

    utt_1 = ds.new_utterance('utt-1', file_1.idx, issuer_idx=issuer_1.idx)
    utt_2 = ds.new_utterance('utt-2', file_2.idx, issuer_idx=issuer_1.idx)
    utt_3 = ds.new_utterance('utt-3', file_3.idx, issuer_idx=issuer_2.idx, start=0, end=1.5)
    utt_4 = ds.new_utterance('utt-4', file_3.idx, issuer_idx=issuer_2.idx, start=1.5, end=2.5)
    utt_5 = ds.new_utterance('utt-5', file_4.idx, issuer_idx=issuer_3.idx)

    utt_1.set_label_list(LabelList(audiomate.corpus.LL_WORD_TRANSCRIPT,
                                   labels=[Label('who am i')]))
    utt_2.set_label_list(LabelList(audiomate.corpus.LL_WORD_TRANSCRIPT,
                                   labels=[Label('who are you', meta={'a': 'hey', 'b': 2})]))
    utt_3.set_label_list(LabelList(audiomate.corpus.LL_WORD_TRANSCRIPT,
                                   labels=[Label('who is he')]))
    utt_4.set_label_list(LabelList(audiomate.corpus.LL_WORD_TRANSCRIPT,
                                   labels=[Label('who are they')]))
    utt_5.set_label_list(LabelList(audiomate.corpus.LL_WORD_TRANSCRIPT,
                                   labels=[Label('who is she')]))

    train_filter = subview.MatchingUtteranceIdxFilter(utterance_idxs={'utt-1', 'utt-2', 'utt-3'})
    sv_train = subview.Subview(ds, filter_criteria=[train_filter])

    dev_filter = subview.MatchingUtteranceIdxFilter(utterance_idxs={'utt-4', 'utt-5'})
    sv_dev = subview.Subview(ds, filter_criteria=[dev_filter])

    ds.import_subview('train', sv_train)
    ds.import_subview('dev', sv_dev)

    ds.new_feature_container('mfcc', '/some/dummy/path')
    ds.new_feature_container('mel', '/some/dummy/path_mel')

    return ds


def create_multi_label_corpus():
    ds = audiomate.Corpus()

    wav_1_path = sample_wav_file('wav_1.wav')
    wav_2_path = sample_wav_file('wav_2.wav')
    wav_3_path = sample_wav_file('wav_3.wav')
    wav_4_path = sample_wav_file('wav_4.wav')

    file_1 = ds.new_file(wav_1_path, track_idx='wav-1')
    file_2 = ds.new_file(wav_2_path, track_idx='wav_2')
    file_3 = ds.new_file(wav_3_path, track_idx='wav_3')
    file_4 = ds.new_file(wav_4_path, track_idx='wav_4')

    issuer_1 = ds.new_issuer('spk-1')
    issuer_2 = ds.new_issuer('spk-2')
    issuer_3 = ds.new_issuer('spk-3')

    utt_1 = ds.new_utterance('utt-1', file_1.idx, issuer_idx=issuer_1.idx)
    utt_2 = ds.new_utterance('utt-2', file_2.idx, issuer_idx=issuer_1.idx)
    utt_3 = ds.new_utterance('utt-3', file_3.idx, issuer_idx=issuer_2.idx, start=0, end=15)
    utt_4 = ds.new_utterance('utt-4', file_3.idx, issuer_idx=issuer_2.idx, start=15, end=25)
    utt_5 = ds.new_utterance('utt-5', file_3.idx, issuer_idx=issuer_2.idx, start=25, end=40)
    utt_6 = ds.new_utterance('utt-6', file_4.idx, issuer_idx=issuer_3.idx, start=0, end=15)
    utt_7 = ds.new_utterance('utt-7', file_4.idx, issuer_idx=issuer_3.idx, start=15, end=25)
    utt_8 = ds.new_utterance('utt-8', file_4.idx, issuer_idx=issuer_3.idx, start=25, end=40)

    utt_1.set_label_list(LabelList(labels=[
        Label('music', 0, 5),
        Label('speech', 5, 12),
        Label('music', 13, 15)
    ]))

    utt_2.set_label_list(LabelList(labels=[
        Label('music', 0, 5),
        Label('speech', 5, 12),
        Label('music', 13, 15)
    ]))

    utt_3.set_label_list(LabelList(labels=[
        Label('music', 0, 1),
        Label('speech', 2, 6)
    ]))

    utt_4.set_label_list(LabelList(labels=[
        Label('music', 0, 5),
        Label('speech', 5, 12),
        Label('music', 13, 15)
    ]))

    utt_5.set_label_list(LabelList(labels=[
        Label('speech', 0, 7)
    ]))

    utt_6.set_label_list(LabelList(labels=[
        Label('music', 0, 5),
        Label('speech', 5, 12),
        Label('music', 13, 15)
    ]))

    utt_7.set_label_list(LabelList(labels=[
        Label('music', 0, 5),
        Label('speech', 5, 11)
    ]))

    utt_8.set_label_list(LabelList(labels=[
        Label('music', 0, 10)
    ]))

    train_filter = subview.MatchingUtteranceIdxFilter(utterance_idxs={'utt-4', 'utt-5', 'utt-6'})
    sv_train = subview.Subview(ds, filter_criteria=[train_filter])

    dev_filter = subview.MatchingUtteranceIdxFilter(utterance_idxs={'utt-7', 'utt-8'})
    sv_dev = subview.Subview(ds, filter_criteria=[dev_filter])

    ds.import_subview('train', sv_train)
    ds.import_subview('dev', sv_dev)

    ds.new_feature_container('mfcc', '/some/dummy/path/secondmfcc')
    ds.new_feature_container('energy', '/some/dummy/path/energy')

    return ds


def create_single_label_corpus():
    ds = audiomate.Corpus()

    wav_1_path = sample_wav_file('wav_1.wav')
    wav_2_path = sample_wav_file('wav_2.wav')
    wav_3_path = sample_wav_file('wav_3.wav')
    wav_4_path = sample_wav_file('wav_4.wav')

    file_1 = ds.new_file(wav_1_path, track_idx='wav-1')
    file_2 = ds.new_file(wav_2_path, track_idx='wav_2')
    file_3 = ds.new_file(wav_3_path, track_idx='wav_3')
    file_4 = ds.new_file(wav_4_path, track_idx='wav_4')

    issuer_1 = ds.new_issuer('spk-1')
    issuer_2 = ds.new_issuer('spk-2')
    issuer_3 = ds.new_issuer('spk-3')

    utt_1 = ds.new_utterance('utt-1', file_1.idx, issuer_idx=issuer_1.idx)
    utt_2 = ds.new_utterance('utt-2', file_2.idx, issuer_idx=issuer_1.idx)
    utt_3 = ds.new_utterance('utt-3', file_3.idx, issuer_idx=issuer_2.idx, start=0, end=15)
    utt_4 = ds.new_utterance('utt-4', file_3.idx, issuer_idx=issuer_2.idx, start=15, end=25)
    utt_5 = ds.new_utterance('utt-5', file_3.idx, issuer_idx=issuer_2.idx, start=25, end=40)
    utt_6 = ds.new_utterance('utt-6', file_4.idx, issuer_idx=issuer_3.idx, start=0, end=15)
    utt_7 = ds.new_utterance('utt-7', file_4.idx, issuer_idx=issuer_3.idx, start=15, end=25)
    utt_8 = ds.new_utterance('utt-8', file_4.idx, issuer_idx=issuer_3.idx, start=25, end=40)

    utt_1.set_label_list(LabelList(labels=[
        Label('music')
    ]))

    utt_2.set_label_list(LabelList(labels=[
        Label('music')
    ]))

    utt_3.set_label_list(LabelList(labels=[
        Label('speech')
    ]))

    utt_4.set_label_list(LabelList(labels=[
        Label('music')
    ]))

    utt_5.set_label_list(LabelList(labels=[
        Label('speech')
    ]))

    utt_6.set_label_list(LabelList(labels=[
        Label('music')
    ]))

    utt_7.set_label_list(LabelList(labels=[
        Label('speech')
    ]))

    utt_8.set_label_list(LabelList(labels=[
        Label('music')
    ]))

    return ds



import collections
import copy

import intervaltree

from .label import Label


class LabelList(object):
    """
    Represents a list of labels which describe an utterance.
    An utterance can have multiple label-lists.

    Args:
        idx (str): An unique identifier for the label-list
                   within a corpus for one utterance.
        labels (list): The list containing the
                       :py:class:`audiomate.annotations.Label`.

    Attributes:
        utterance (Utterance): The utterance this label-list is belonging to.
        label_tree (IntervalTree): The interval-tree storing the labels.

    Example:
        >>> label_list = LabelList(idx='transcription', labels=[
        >>>     Label('this', 0, 2),
        >>>     Label('is', 2, 4),
        >>>     Label('timmy', 4, 8)
        >>> ])
    """

    __slots__ = ['idx', 'label_tree', 'utterance']

    def __init__(self, idx='default', labels=None):
        self.idx = idx
        self.utterance = None

        self.label_tree = intervaltree.IntervalTree()

        if labels is not None:
            self.update(labels)

    def __eq__(self, other):
        data_this = (self.idx, self.label_tree)
        data_other = (other.idx, other.label_tree)
        return data_this == data_other

    def __iter__(self):
        for interval in self.label_tree:
            yield interval.data

    def __len__(self):
        return self.label_tree.__len__()

    def __copy__(self):
        # utterance is ignored intentionally,
        # since it is kind of a weak ref
        return LabelList(
            idx=self.idx,
            labels=[iv.data for iv in self.label_tree]
        )

    def __deepcopy__(self, memo):
        # utterance is ignored intentionally,
        # since it is kind of a weak ref
        return LabelList(
            idx=self.idx,
            labels=copy.deepcopy([iv.data for iv in self.label_tree], memo)
        )

    @property
    def labels(self):
        """ Return list of labels. """
        return list(self)

    @property
    def start(self):
        """ Return start of the earliest starting label (lower bound). """
        return self.label_tree.begin()

    @property
    def end(self):
        """ Return end of the lastly ending label (upper bound). """
        return self.label_tree.end()

    @property
    def total_length(self):
        """
        Return the cumulative length of all labels. (Number of characters)
        """
        return sum([label.length for label in self.labels])

    #
    #   Alteration
    #

    def add(self, label):
        """
        Add a label to the end of the list.

        Args:
            label (Label): The label to add.
        """
        label.label_list = self
        self.label_tree.addi(label.start, label.end, label)

    def addl(self, value, start=0.0, end=float('inf')):
        """
        Shortcut for ``add(Label(value, start, end))``.
        """
        self.add(Label(value, start=start, end=end))

    def update(self, labels):
        """
        Add a list of labels to the end of the list.

        Args:
            labels (list): Labels to add.
        """
        ivs = []

        for label in labels:
            label.label_list = self
            ivs.append(intervaltree.Interval(label.start, label.end, label))

        self.label_tree.update(ivs)

    def apply(self, fn):
        """
        Apply the given function `fn` to every label in this label list.
        `fn` is a function of one argument that receives the current label
        which can then be edited in place.

        Args:
            fn (func): Function to apply to every label

        Example:
            >>> ll = LabelList(labels=[
            ...     Label('a_label', 1.0, 2.0),
            ...     Label('another_label', 2.0, 3.0)
            ... ])
            >>> def shift_labels(label):
            ...     label.start += 1.0
            ...     label.end += 1.0
            ...
            >>> ll.apply(shift_labels)
            >>> ll.labels
            [Label(a_label, 2.0, 3.0), Label(another_label, 3.0, 4.0)]
        """
        for label in self.labels:
            fn(label)

    def merge_overlaps(self, threshold=0.0):
        """
        Merge overlapping labels with the same value.
        Two labels are considered overlapping,
        if ``l2.start - l1.end < threshold``.

        Args:
            threshold (float): Maximal distance between two labels
                               to be considered as overlapping.
                               (default: 0.0)

        Example:
            >>> ll = LabelList(labels=[
            ...     Label('a_label', 1.0, 2.0),
            ...     Label('a_label', 1.5, 2.7),
            ...     Label('b_label', 1.0, 2.0),
            ... ])
            >>> ll.merge_overlapping_labels()
            >>> ll.labels
            [
                Label('a_label', 1.0, 2.7),
                Label('b_label', 1.0, 2.0),
            ]
        """

        updated_labels = []
        all_intervals = self.label_tree.copy()

        # recursivly find a group of overlapping labels with the same value
        def recursive_overlaps(interval):
            range_start = interval.begin - threshold
            range_end = interval.end + threshold

            direct_overlaps = all_intervals.overlap(range_start, range_end)
            all_overlaps = [interval]
            all_intervals.discard(interval)

            for overlap in direct_overlaps:
                if overlap.data.value == interval.data.value:
                    all_overlaps.extend(recursive_overlaps(overlap))

            return all_overlaps

        # For every remaining interval
        # - Find overlapping intervals recursively
        # - Remove them
        # - Create a concatenated new label
        while not all_intervals.is_empty():
            next_interval = list(all_intervals)[0]
            overlapping = recursive_overlaps(next_interval)

            ov_start = float('inf')
            ov_end = 0.0
            ov_value = next_interval.data.value

            for overlap in overlapping:
                ov_start = min(ov_start, overlap.begin)
                ov_end = max(ov_end, overlap.end)
                all_intervals.discard(overlap)

            updated_labels.append(Label(
                ov_value,
                ov_start,
                ov_end
            ))

        # Replace the old labels with the updated ones
        self.label_tree.clear()
        self.update(updated_labels)

    #
    #   Statistics
    #

    def label_total_duration(self):
        """
        Return for each distinct label value the total duration of all occurrences.

        Returns:
            dict: A dictionary containing for every label-value (key)
                  the total duration in seconds (value).

        Example:
            >>> ll = LabelList(labels=[
            >>>     Label('a', 3, 5),
            >>>     Label('b', 5, 8),
            >>>     Label('a', 8, 10),
            >>>     Label('b', 10, 14),
            >>>     Label('a', 15, 18.5)
            >>> ])
            >>> ll.label_total_duration()
            {'a': 7.5 'b': 7.0}
        """

        durations = collections.defaultdict(float)

        for label in self:
            durations[label.value] += label.duration

        return durations

    def label_values(self):
        """
        Return a list of all occuring label values.

        Returns:
            list: Lexicographically sorted list (str) of label values.

        Example:
            >>> ll = LabelList(labels=[
            >>>     Label('a', 3.2, 4.5),
            >>>     Label('b', 5.1, 8.9),
            >>>     Label('c', 7.2, 10.5),
            >>>     Label('d', 10.5, 14),
            >>>     Label('d', 15, 18)
            >>> ])
            >>> ll.label_values()
            ['a', 'b', 'c', 'd']
        """

        all_labels = set([l.value for l in self])
        return sorted(all_labels)

    def label_count(self):
        """
        Return for each label the number of occurrences within the list.

        Returns:
            dict: A dictionary containing for every label-value (key)
            the number of occurrences (value).

        Example:
            >>> ll = LabelList(labels=[
            >>>     Label('a', 3.2, 4.5),
            >>>     Label('b', 5.1, 8.9),
            >>>     Label('a', 7.2, 10.5),
            >>>     Label('b', 10.5, 14),
            >>>     Label('a', 15, 18)
            >>> ])
            >>> ll.label_count()
            {'a': 3 'b': 2}
        """

        occurrences = collections.defaultdict(int)

        for label in self:
            occurrences[label.value] += 1

        return occurrences

    def all_tokens(self, delimiter=' '):
        """
        Return a list of all tokens occurring in the label-list.

        Args:
            delimiter (str): The delimiter used to split labels into tokens
                             (see :meth:`audiomate.annotations.Label.tokenized`).

        Returns:
             :class:`set`: A set of distinct tokens.
        """
        tokens = set()

        for label in self:
            tokens = tokens.union(set(label.tokenized(delimiter=delimiter)))

        return tokens

    #
    #   Query Label Values
    #

    def join(self, delimiter=' ', overlap_threshold=0.1):
        """
        Return a string with all labels concatenated together.
        The order of the labels is defined by the start of the label.
        If the overlapping between two labels is greater than ``overlap_threshold``,
        an Exception is thrown.

        Args:
            delimiter (str): A string to join two consecutive labels.
            overlap_threshold (float): Maximum overlap between two consecutive labels.

        Returns:
            str: A string with all labels concatenated together.

        Example:
            >>> ll = LabelList(idx='some', labels=[
            >>>     Label('a', start=0, end=4),
            >>>     Label('b', start=3.95, end=6.0),
            >>>     Label('c', start=7.0, end=10.2),
            >>>     Label('d', start=10.3, end=14.0)
            >>> ])
            >>> ll.join(' - ')
            'a - b - c - d'
        """

        sorted_by_start = sorted(self.labels)
        concat_values = []
        last_label_end = None

        for label in sorted_by_start:
            if last_label_end is None or (last_label_end - label.start < overlap_threshold and last_label_end > 0):
                concat_values.append(label.value)
                last_label_end = label.end
            else:
                raise ValueError('Labels overlap, not able to define the correct order')

        return delimiter.join(concat_values)

    def tokenized(self, delimiter=' ', overlap_threshold=0.1):
        """
        Return a ordered list of tokens based on all labels.
        Joins all token from all labels (``label.tokenized()```).
        If the overlapping between two labels is greater than ``overlap_threshold``,
        an Exception is thrown.

        Args:
            delimiter (str): The delimiter used to split labels into tokens. (default: space)
            overlap_threshold (float): Maximum overlap between two consecutive labels.

        Returns:
            str: A list containing tokens of all labels ordered according to the label order.

        Example:
            >>> ll = LabelList(idx='some', labels=[
            >>>     Label('a d q', start=0, end=4),
            >>>     Label('b', start=3.95, end=6.0),
            >>>     Label('c a', start=7.0, end=10.2),
            >>>     Label('f g', start=10.3, end=14.0)
            >>> ])
            >>> ll.tokenized(delimiter=' ', overlap_threshold=0.1)
            ['a', 'd', 'q', 'b', 'c', 'a', 'f', 'g']
        """

        sorted_by_start = sorted(self.labels)
        tokens = []
        last_label_end = None

        for label in sorted_by_start:
            if last_label_end is None or (last_label_end - label.start < overlap_threshold and last_label_end > 0):
                tokens.extend(label.tokenized(delimiter=delimiter))
                last_label_end = label.end
            else:
                raise ValueError('Labels overlap, not able to define the correct order')

        return tokens

    #
    #   Restructuring
    #

    def separated(self):
        """
        Create a separate Label-List for every distinct label-value.

        Returns:
            dict: A dictionary with distinct label-values as keys.
                  Every value is a LabelList containing only labels with the same value.

        Example:
            >>> ll = LabelList(idx='some', labels=[
            >>>     Label('a', start=0, end=4),
            >>>     Label('b', start=3.95, end=6.0),
            >>>     Label('a', start=7.0, end=10.2),
            >>>     Label('b', start=10.3, end=14.0)
            >>> ])
            >>> s = ll.separate()
            >>> s['a'].labels
            [Label('a', start=0, end=4), Label('a', start=7.0, end=10.2)]
            >>> s['b'].labels
            [Label('b', start=3.95, end=6.0), Label('b', start=10.3, end=14.0)]
        """
        separated_lls = collections.defaultdict(LabelList)

        for label in self.labels:
            separated_lls[label.value].add(label)

        for ll in separated_lls.values():
            ll.idx = self.idx

        return separated_lls

    def labels_in_range(self, start, end, fully_included=False):
        """
        Return a list of labels, that are within the given range.
        Also labels that only overlap are included.

        Args:
            start(float): Start-time in seconds.
            end(float): End-time in seconds.
            fully_included(bool): If ``True``, only labels fully included
                                   in the range are returned. Otherwise
                                   also overlapping ones are returned.
                                   (default ``False``)

        Returns:
            list: List of labels in the range.

        Example:
            >>> ll = LabelList(labels=[
            >>>     Label('a', 3.2, 4.5),
            >>>     Label('b', 5.1, 8.9),
            >>>     Label('c', 7.2, 10.5),
            >>>     Label('d', 10.5, 14)
            >>>])
            >>> ll.labels_in_range(6.2, 10.1)
            [Label('b', 5.1, 8.9), Label('c', 7.2, 10.5)]
        """

        if fully_included:
            intervals = self.label_tree.envelop(start, end)
        else:
            intervals = self.label_tree.overlap(start, end)

        return [iv.data for iv in intervals]

    def ranges(self, yield_ranges_without_labels=False, include_labels=None):
        """
        Generate all ranges of the label-list. A range is defined
        as a part of the label-list for which the same labels are defined.

        Args:
            yield_ranges_without_labels(bool): If True also yields ranges for
                                                which no labels are defined.
            include_labels(list): If not empty, only the label values in
                                   the list will be considered.

        Returns:
            generator: A generator which yields one range
            (tuple start/end/list-of-labels) at a time.

        Example:
            >>> ll = LabelList(labels=[
            >>>     Label('a', 3.2, 4.5),
            >>>     Label('b', 5.1, 8.9),
            >>>     Label('c', 7.2, 10.5),
            >>>     Label('d', 10.5, 14)
            >>>])
            >>> ranges = ll.ranges()
            >>> next(ranges)
            (3.2, 4.5, [ < audiomate.annotations.Label at 0x1090527c8 > ])
            >>> next(ranges)
            (4.5, 5.1, [])
            >>> next(ranges)
            (5.1, 7.2, [ < audiomate.annotations.label.Label at 0x1090484c8 > ])
        """

        tree_copy = self.label_tree.copy()

        # Remove labels not included
        if include_labels is not None:
            for iv in list(tree_copy):
                if iv.data.value not in include_labels:
                    tree_copy.remove(iv)

        def reduce(x, y):
            x.append(y)
            return x

        # Split labels when overlapping and merge equal ranges to a list of labels
        tree_copy.split_overlaps()
        tree_copy.merge_equals(data_reducer=reduce, data_initializer=[])

        intervals = sorted(tree_copy)
        last_end = intervals[0].begin

        # yield range by range
        for i in range(len(intervals)):
            iv = intervals[i]

            # yield an empty range if necessary
            if yield_ranges_without_labels and iv.begin > last_end:
                yield (last_end, iv.begin, [])

            yield (iv.begin, iv.end, iv.data)

            last_end = iv.end

    def split(self, cutting_points, shift_times=False, overlap=0.0):
        """
        Split the label-list into x parts and return them as new label-lists.
        x is defined by the number of cutting-points(``x == len(cutting_points) + 1``)

        The result is a list of label-lists corresponding to each part.
        Label-list 0 contains labels between ``0`` and ``cutting_points[0]``.
        Label-list 1 contains labels between ``cutting_points[0]`` and ``cutting_points[1]``.
        And so on.

        Args:
            cutting_points(list): List of floats defining the points in seconds,
                                  where the label-list is splitted.
            shift_times(bool): If True, start and end-time are shifted in splitted label-lists.
                               So the start is relative to the cutting point and
                               not to the beginning of the original label-list.
            overlap(float): Amount of overlap in seconds. This amount is subtracted
                            from a start-cutting-point, and added to a end-cutting-point.

        Returns:
            list: A list of of: class: `audiomate.annotations.LabelList`.

        Example:

            >>> ll = LabelList(labels=[
            >>>     Label('a', 0, 5),
            >>>     Label('b', 5, 10),
            >>>     Label('c', 11, 15),
            >>>])
            >>>
            >>> res = ll.split([4.1, 8.9, 12.0])
            >>> len(res)
            4
            >>> res[0].labels
            [Label('a', 0.0, 4.1)]
            >>> res[1].labels
            [
                Label('a', 4.1, 5.0),
                Label('b', 5.0, 8.9)
            ]
            >>> res[2].labels
            [
                Label('b', 8.9, 10.0),
                Label('c', 11.0, 12.0)
            ]
            >>> res[3].labels
            [Label('c', 12.0, 15.0)]

        If ``shift_times = True``, the times are adjusted to be relative
        to the cutting-points for every label-list but the first.

            >>> ll = LabelList(labels=[
            >>>     Label('a', 0, 5),
            >>>     Label('b', 5, 10),
            >>>])
            >>>
            >>> res = ll.split([4.6])
            >>> len(res)
            4
            >>> res[0].labels
            [Label('a', 0.0, 4.6)]
            >>> res[1].labels
            [
                Label('a', 0.0, 0.4),
                Label('b', 0.4, 5.4)
            ]
        """

        if len(cutting_points) == 0:
            raise ValueError('At least one cutting-point is needed!')

        # we have to loop in sorted order
        cutting_points = sorted(cutting_points)

        splits = []
        iv_start = 0.0

        for i in range(len(cutting_points) + 1):
            if i < len(cutting_points):
                iv_end = cutting_points[i]
            else:
                iv_end = float('inf')

            # get all intervals intersecting range
            intervals = self.label_tree.overlap(
                iv_start - overlap,
                iv_end + overlap
            )

            cp_splits = LabelList(idx=self.idx)

            # Extract labels from intervals with updated times
            for iv in intervals:
                label = copy.deepcopy(iv.data)
                label.start = max(0, iv_start - overlap, label.start)
                label.end = min(iv_end + overlap, label.end)

                if shift_times:
                    orig_start = max(0, iv_start - overlap)
                    label.start -= orig_start
                    label.end -= orig_start

                cp_splits.add(label)

            splits.append(cp_splits)
            iv_start = iv_end

        return splits

    #
    #   Convenience Constructors
    #

    @classmethod
    def create_single(cls, value, idx='default'):
        """ Create a label-list with a single label containing the given value. """

        return LabelList(idx=idx, labels=[
            Label(value=value)
        ])

    @classmethod
    def with_label_values(cls, values, idx='default'):
        """
        Create a new label-list containing labels with the given values.
        All labels will have default start/end values of 0 and ``inf``.

        Args:
            values(list): List of values(str) that should be created and appended
                           to the label-list.
            idx(str): The idx of the label-list.

        Returns:
            (LabelList): New label-list.

        Example:
            >>> ll = LabelList.with_label_values(['a', 'x', 'z'], idx='letters')
            >>> ll.idx
            'letters'
            >>> ll.labels
            [
                Label('a', 0, inf),
                Label('x', 0, inf),
                Label('z', 0, inf),
            ]
        """

        ll = LabelList(idx=idx)

        for label_value in values:
            ll.add(Label(label_value))

        return ll



import os
import collections

import numpy as np
import scipy

import audiomate
from . import base
from audiomate.utils import textfile


class MozillaDeepSpeechWriter(base.CorpusWriter):
    """
    Writes files to use for training with Mozilla DeepSpeech (https://github.com/mozilla/DeepSpeech).

    Since it is expected that every utterance is in a separate file,
    any utterances that are not in separate file in the original corpus,
    are extracted into a separate file in the subfolder `audio` of the target path.

    Args:
        transcription_label_list_idx (str): The transcriptions are used from the label-list with this id.
    """

    def __init__(self, transcription_label_list_idx=audiomate.corpus.LL_WORD_TRANSCRIPT):
        self.transcription_label_list_idx = transcription_label_list_idx

    @classmethod
    def type(cls):
        return 'mozilla-deepspeech'

    def _save(self, corpus, path):
        records = []
        subset_utterance_ids = {idx: list(subset.utterances.keys()) for idx, subset in corpus.subviews.items()}
        subset_records = collections.defaultdict(list)

        audio_folder = os.path.join(path, 'audio')
        os.makedirs(audio_folder, exist_ok=True)

        for utterance_idx in sorted(corpus.utterances.keys()):
            utterance = corpus.utterances[utterance_idx]

            if utterance.start == 0 and utterance.end == float('inf'):
                audio_path = utterance.track.path
            else:
                audio_path = os.path.join(audio_folder, '{}.wav'.format(utterance.idx))
                sampling_rate = utterance.sampling_rate
                data = utterance.read_samples()

                data = (data * 32768).astype(np.int16)

                scipy.io.wavfile.write(audio_path, sampling_rate, data)

            size = os.stat(audio_path).st_size
            transcript = utterance.label_lists[self.transcription_label_list_idx].join()

            # Add to the full list
            record = [audio_path, size, transcript]
            records.append(record)

            # Check / Add to subview lists
            for subset_idx, utt_ids in subset_utterance_ids.items():
                if utterance_idx in utt_ids:
                    subset_records[subset_idx].append(record)

        # Write full list
        records.insert(0, ['wav_filename', 'wav_filesize', 'transcript'])
        records_path = os.path.join(path, 'all.csv')
        textfile.write_separated_lines(records_path, records, separator=',', sort_by_column=-1)

        # Write subset lists
        for subset_idx, records in subset_records.items():
            if len(records) > 0:
                records.insert(0, ['wav_filename', 'wav_filesize', 'transcript'])
                subset_file_path = os.path.join(path, '{}.csv'.format(subset_idx))
                textfile.write_separated_lines(subset_file_path, records, separator=',', sort_by_column=-1)



import abc

import librosa
import numpy as np

from audiomate import containers
from audiomate.utils import units


class Processor(metaclass=abc.ABCMeta):
    """
    The processor base class provides the functionality to process audio data on different levels
    (Corpus, Utterance, Track). For every level there is an offline and an online method.
    In the offline mode the data is processed in one step (e.g. the whole track/utterance at once).
    This means the ``process_frames`` method is called with all the frames of the track/utterance.
    In online mode the data is processed in chunks, so the ``process_frames`` method is called multiple times
    per track/utterance with different chunks.

    To implement a concrete processor the ``process_frames`` method has to be implemented.
    This method is called in online and offline mode. So it is up to the user to determine
    if a processor can be called in either online or offline mode, maybe both. This differs between use cases.

    If the implementation of a processor does change the frame or hop-size,
    it is expected to provide a transform via the ``frame_transform`` method.
    Frame-size and hop-size are measured in samples regarding the original audio signal (or simply its sampling rate).
    """

    def process_corpus(self, corpus, output_path, frame_size=400, hop_size=160, sr=None):
        """
        Process all utterances of the given corpus and save the processed features in a feature-container.
        The utterances are processed in **offline** mode so the full utterance in one go.

        Args:
            corpus (Corpus): The corpus to process the utterances from.
            output_path (str): A path to save the feature-container to.
            frame_size (int): The number of samples per frame.
            hop_size (int): The number of samples between two frames.
            sr (int): Use the given sampling rate. If None uses the native sampling rate from the underlying data.

        Returns:
            FeatureContainer: The feature-container containing the processed features.
        """

        def processing_func(utterance, feat_container, frame_size, hop_size, sr, corpus):
            data = self.process_utterance(utterance, frame_size=frame_size, hop_size=hop_size, sr=sr, corpus=corpus)
            feat_container.set(utterance.idx, data)

        return self._process_corpus(corpus, output_path, processing_func,
                                    frame_size=frame_size, hop_size=hop_size, sr=sr)

    def process_corpus_online(self, corpus, output_path, frame_size=400, hop_size=160,
                              chunk_size=1, buffer_size=5760000):
        """
        Process all utterances of the given corpus and save the processed features in a feature-container.
        The utterances are processed in **online** mode, so chunk by chunk.

        Args:
            corpus (Corpus): The corpus to process the utterances from.
            output_path (str): A path to save the feature-container to.
            frame_size (int): The number of samples per frame.
            hop_size (int): The number of samples between two frames.
            chunk_size (int): Number of frames to process per chunk.
            buffer_size (int): Number of samples to load into memory at once.
                             The exact number of loaded samples depends on the block-size of the audioread library.
                             So it can be of block-size higher, where the block-size is typically 1024 or 4096.

        Returns:
            FeatureContainer: The feature-container containing the processed features.
        """

        def processing_func(utterance, feat_container, frame_size, hop_size, corpus, sr):
            for chunk in self.process_utterance_online(utterance,
                                                       frame_size=frame_size,
                                                       hop_size=hop_size,
                                                       corpus=corpus,
                                                       chunk_size=chunk_size,
                                                       buffer_size=buffer_size):
                feat_container.append(utterance.idx, chunk)

        return self._process_corpus(corpus, output_path, processing_func,
                                    frame_size=frame_size, hop_size=hop_size, sr=None)

    def process_features(self, corpus, input_features, output_path):
        """
        Process all features of the given corpus and save the processed features in a feature-container.
        The features are processed in **offline** mode, all features of an utterance at once.

        Args:
            corpus (Corpus): The corpus to process the utterances from.
            input_features (FeatureContainer): The feature-container to process the frames from.
            output_path (str): A path to save the feature-container to.

        Returns:
            FeatureContainer: The feature-container containing the processed features.
        """
        feat_container = containers.FeatureContainer(output_path)
        feat_container.open()

        input_features.open()

        for utterance in corpus.utterances.values():
            sampling_rate = input_features.sampling_rate
            frames = input_features.get(utterance.idx, mem_map=False)
            processed = self.process_frames(frames, sampling_rate, offset=0, last=True,
                                            utterance=utterance, corpus=corpus)
            feat_container.set(utterance.idx, processed)

        tf_frame_size, tf_hop_size = self.frame_transform(input_features.frame_size, input_features.hop_size)
        feat_container.frame_size = tf_frame_size
        feat_container.hop_size = tf_hop_size
        feat_container.sampling_rate = input_features.sampling_rate

        feat_container.close()

        return feat_container

    def process_features_online(self, corpus, input_features, output_path, chunk_size=1):
        """
        Process all features of the given corpus and save the processed features in a feature-container.
        The features are processed in **online** mode, chunk by chunk.

        Args:
            corpus (Corpus): The corpus to process the utterances from.
            input_features (FeatureContainer): The feature-container to process the frames from.
            output_path (str): A path to save the feature-container to.
            chunk_size (int): Number of frames to process per chunk.

        Returns:
            FeatureContainer: The feature-container containing the processed features.
        """
        feat_container = containers.FeatureContainer(output_path)
        feat_container.open()

        input_features.open()

        for utterance in corpus.utterances.values():
            sampling_rate = input_features.sampling_rate
            frames = input_features.get(utterance.idx, mem_map=True)

            current_frame = 0

            while current_frame < frames.shape[0]:
                last = current_frame + chunk_size > frames.shape[0]
                to_frame = current_frame + chunk_size

                chunk = frames[current_frame:to_frame]

                processed = self.process_frames(chunk, sampling_rate, current_frame,
                                                last=last, utterance=utterance, corpus=corpus)

                if processed is not None:
                    feat_container.append(utterance.idx, processed)

                current_frame += chunk_size

        tf_frame_size, tf_hop_size = self.frame_transform(input_features.frame_size, input_features.hop_size)
        feat_container.frame_size = tf_frame_size
        feat_container.hop_size = tf_hop_size
        feat_container.sampling_rate = input_features.sampling_rate

        feat_container.close()

        return feat_container

    def process_utterance(self, utterance, frame_size=400, hop_size=160, sr=None, corpus=None):
        """
        Process the utterance in **offline** mode, in one go.

        Args:
            utterance (Utterance): The utterance to process.
            frame_size (int): The number of samples per frame.
            hop_size (int): The number of samples between two frames.
            sr (int): Use the given sampling rate. If None uses the native sampling rate from the underlying data.
            corpus (Corpus): The corpus this utterance is part of, if available.

        Returns:
            np.ndarray: The processed features.
        """
        return self.process_track(utterance.track, frame_size=frame_size, hop_size=hop_size, sr=sr,
                                  start=utterance.start, end=utterance.end, utterance=utterance, corpus=corpus)

    def process_utterance_online(self, utterance, frame_size=400, hop_size=160, chunk_size=1,
                                 buffer_size=5760000, corpus=None):
        """
        Process the utterance in **online** mode, chunk by chunk.
        The processed chunks are yielded one after another.

        Args:
            utterance (Utterance): The utterance to process.
            frame_size (int): The number of samples per frame.
            hop_size (int): The number of samples between two frames.
            chunk_size (int): Number of frames to process per chunk.
            buffer_size (int): Number of samples to load into memory at once.
                             The exact number of loaded samples depends on the block-size of the audioread library.
                             So it can be of block-size higher, where the block-size is typically 1024 or 4096.
            corpus (Corpus): The corpus this utterance is part of, if available.

        Returns:
            Generator: A generator that yield processed chunks.
        """
        return self.process_track_online(utterance.track,
                                         frame_size=frame_size,
                                         hop_size=hop_size,
                                         start=utterance.start,
                                         end=utterance.end,
                                         utterance=utterance,
                                         corpus=corpus,
                                         chunk_size=chunk_size,
                                         buffer_size=buffer_size)

    def process_track(self, track, frame_size=400, hop_size=160, sr=None,
                      start=0, end=float('inf'), utterance=None, corpus=None):
        """
        Process the track in **offline** mode, in one go.

        Args:
            track (Track): The track to process.
            frame_size (int): The number of samples per frame.
            hop_size (int): The number of samples between two frames.
            sr (int): Use the given sampling rate. If ``None``,
                      uses the native sampling rate from the underlying data.
            start (float): The point within the track in seconds,
                           to start processing from.
            end (float): The point within the track in seconds,
                         to end processing.
            utterance (Utterance): The utterance that is associated with
                                   this track, if available.
            corpus (Corpus): The corpus this track is part of, if available.

        Returns:
            np.ndarray: The processed features.
        """
        frame_settings = units.FrameSettings(frame_size, hop_size)

        if end != float('inf'):
            samples = track.read_samples(sr=sr, offset=start, duration=end-start)
        else:
            samples = track.read_samples(sr=sr, offset=start)

        if sr is None:
            sr = track.sampling_rate

        if samples.size <= 0:
            raise ValueError('Track {} has no samples'.format(track.idx))

        # Pad with zeros to match frames
        num_frames = frame_settings.num_frames(samples.size)
        num_pad_samples = (num_frames - 1) * hop_size + frame_size

        if num_pad_samples > samples.size:
            samples = np.pad(samples, (0, num_pad_samples - samples.size), mode='constant', constant_values=0)

        # Get sampling-rate if not given
        sampling_rate = sr or utterance.sampling_rate

        frames = librosa.util.frame(samples, frame_length=frame_size, hop_length=hop_size).T
        return self.process_frames(frames, sampling_rate, 0, last=True, utterance=utterance, corpus=corpus)

    def process_track_online(self, track, frame_size=400, hop_size=160,
                             start=0, end=float('inf'), utterance=None, corpus=None,
                             chunk_size=1, buffer_size=5760000):
        """
        Process the track in **online** mode, chunk by chunk.
        The processed chunks are yielded one after another.

        Args:
            track (Track): The track to process.
            frame_size (int): The number of samples per frame.
            hop_size (int): The number of samples between two frames.
            start (float): The point within the track in seconds to start processing from.
            end (float): The point within the trac in seconds to end processing.
            utterance (Utterance): The utterance that is associated with this track, if available.
            corpus (Corpus): The corpus this track is part of, if available.
            chunk_size (int): Number of frames to process per chunk.
            buffer_size (int): Number of samples to load into memory at once.
                               The exact number of loaded samples depends
                               on the type of track.
                               It can be of block-size higher,
                               where the block-size is typically 1024 or 4096.

        Returns:
            Generator: A generator that yield processed chunks.
        """

        current_frame = 0
        frames = []

        duration = None
        sr = track.sampling_rate

        if end != float('inf'):
            duration = end - start

        # Process chunks that are within end bounds
        for frame, is_last in track.read_frames(frame_size,
                                                hop_size,
                                                offset=start,
                                                duration=duration,
                                                buffer_size=buffer_size):

            frames.append(frame)

            if len(frames) == chunk_size:
                processed = self.process_frames(np.array(frames), sr, current_frame,
                                                last=is_last, utterance=utterance, corpus=corpus)
                if processed is not None:
                    yield processed
                current_frame += chunk_size
                frames = frames[chunk_size:]

        # Process overlapping chunks with zero frames at the end
        if len(frames) > 0:
            processed = self.process_frames(np.array(frames), sr, current_frame,
                                            last=True, utterance=utterance, corpus=corpus)
            yield processed

    @abc.abstractmethod
    def process_frames(self, data, sampling_rate, offset=0, last=False, utterance=None, corpus=None):
        """
        Process the given chunk of frames. Depending on online or offline mode,
        the given chunk is either the full data or just part of it.

        Args:
            data (np.ndarray): nD Array of frames (num-frames x frame-dimensions).
            sampling_rate (int): The sampling rate of the underlying signal.
            offset (int): The index of the first frame in the chunk. In offline mode always 0.
                          (Relative to the first frame of the utterance/sequence)
            last (bool): True indicates that this is the last frame of the sequence/utterance.
                         In offline mode always True.
            utterance (Utterance): The utterance the frame is from, if available.
            corpus (Corpus): The corpus the frame is from, if available.

        Returns:
            np.ndarray: The processed frames.
        """
        pass

    def frame_transform(self, frame_size, hop_size):
        """
        If the processor changes the number of samples that build up a frame or
        the number of samples between two consecutive frames (hop-size),
        this function needs transform the original frame- and/or hop-size.

        This is used to store the frame-size and hop-size in a feature-container.
        In the end one can calculate start and end time of a frame with this information.

        By default it is assumed that the processor doesn't change the frame-size and the hop-size.

        Args:
            frame_size (int): The original frame-size.
            hop_size (int): The original hop-size.

        Returns:
            tuple: The (frame-size, hop-size) after processing.
        """
        return frame_size, hop_size

    def _process_corpus(self, corpus, output_path, processing_func, frame_size=400, hop_size=160, sr=None):
        """ Utility function for processing a corpus with a separate processing function. """
        feat_container = containers.FeatureContainer(output_path)
        feat_container.open()

        sampling_rate = -1

        for utterance in corpus.utterances.values():
            utt_sampling_rate = utterance.sampling_rate

            if sr is None:
                if sampling_rate > 0 and sampling_rate != utt_sampling_rate:
                    raise ValueError(
                        'File {} has a different sampling-rate than the previous ones!'.format(utterance.track.idx))

                sampling_rate = utt_sampling_rate

            processing_func(utterance, feat_container, frame_size, hop_size, sr, corpus)

        tf_frame_size, tf_hop_size = self.frame_transform(frame_size, hop_size)
        feat_container.frame_size = tf_frame_size
        feat_container.hop_size = tf_hop_size
        feat_container.sampling_rate = sr or sampling_rate

        feat_container.close()

        return feat_container



import numpy as np

from audiomate.utils import stats
from . import container


class FeatureContainer(container.Container):
    """
    The FeatureContainer is a container for storing features
    extracted from audio data. Features are array-like data,
    where every feature represents the properties of a given segment of audio.

    Args:
        path (str): Path to where the HDF5 file is stored.
                    If the file doesn't exist, one is created.
        mode (str): Either 'r' for read-only, 'w' for truncate and write or
                    'a' for append. (default: 'a').

    Example:
        >>> fc = FeatureContainer('/path/to/hdf5file')
        >>> with fc:
        >>>     fc.set('utt-1', np.array([1,2,3,4]))
        >>>     data = fc.get('utt-1')
        array([1, 2, 3, 4])
    """

    @property
    def frame_size(self):
        """ The number of samples used per frame. """
        self.raise_error_if_not_open()
        return self._file.attrs['frame-size']

    @frame_size.setter
    def frame_size(self, frame_size):
        self.raise_error_if_not_open()
        self._file.attrs['frame-size'] = frame_size

    @property
    def hop_size(self):
        """ The number of samples between two frames. """
        self.raise_error_if_not_open()
        return self._file.attrs['hop-size']

    @hop_size.setter
    def hop_size(self, hop_size):
        self.raise_error_if_not_open()
        self._file.attrs['hop-size'] = hop_size

    @property
    def sampling_rate(self):
        """ The sampling-rate of the signal these frames are based on. """
        self.raise_error_if_not_open()
        return self._file.attrs['sampling-rate']

    @sampling_rate.setter
    def sampling_rate(self, sampling_rate):
        self.raise_error_if_not_open()
        self._file.attrs['sampling-rate'] = sampling_rate

    def stats(self):
        """
        Return statistics calculated overall features in the container.

        Note:
            The feature container has to be opened in advance.

        Returns:
            DataStats: Statistics overall data points of all features.
        """
        self.raise_error_if_not_open()

        per_key_stats = self.stats_per_key()

        return stats.DataStats.concatenate(per_key_stats.values())

    def stats_per_key(self):
        """
        Return statistics calculated for each key in the container.

        Note:
            The feature container has to be opened in advance.

        Returns:
            dict: A dictionary containing a DataStats object for each key.
        """
        self.raise_error_if_not_open()

        all_stats = {}

        for key, data in self._file.items():
            data = data[()]
            all_stats[key] = stats.DataStats(float(np.mean(data)),
                                             float(np.var(data)),
                                             np.min(data),
                                             np.max(data),
                                             data.size)

        return all_stats



import random

import pytest
from unittest import mock

from audiomate import annotations
from audiomate.corpus.subset import splitting
from tests import resources


INITIAL_SEED = 99


@pytest.fixture
def corpus():
    return resources.create_multi_label_corpus()


@pytest.fixture
def splitter():
    corpus = resources.create_multi_label_corpus()
    return splitting.Splitter(corpus, random_seed=INITIAL_SEED)


class TestSplitter:

    #
    # SPLIT()
    #

    @mock.patch('audiomate.corpus.subset.utils.split_identifiers')
    def test_split(self, split_mock, splitter):
        split_mock.return_value = {
            'train': ['utt-1', 'utt-3'],
            'test': ['utt-3', 'utt-4']
        }
        res = splitter.split({'train': 0.5, 'test': 0.5})

        assert res['train'].utterances.keys() == {'utt-1', 'utt-3'}
        assert res['test'].utterances.keys() == {'utt-3', 'utt-4'}

        split_mock.assert_called_with(
            list(splitter.corpus.utterances.keys()),
            {'train': 0.5, 'test': 0.5},
            seed=mock.ANY
        )

    @mock.patch('audiomate.corpus.subset.utils.split_identifiers')
    def test_split_passes_seed(self, split_mock, splitter):
        splitter.split(
            {'train': 0.5, 'test': 0.5},
            separate_issuers=False
        )

        r = random.Random(INITIAL_SEED)

        split_mock.assert_called_with(
            mock.ANY,
            mock.ANY,
            seed=r.random()
        )

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_issuer_separated(self, split_mock, splitter):
        split_mock.return_value = {
            'train': ['spk-1', 'spk-2'],
            'test': ['spk-3']
        }
        res = splitter.split(
            {'train': 0.5, 'test': 0.5},
            separate_issuers=True
        )

        assert res['train'].utterances.keys() == {
            'utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'
        }
        assert res['test'].utterances.keys() == {
            'utt-6', 'utt-7', 'utt-8'
        }

        split_mock.assert_called_with(
            {
                'spk-1': {'count': 2},
                'spk-2': {'count': 3},
                'spk-3': {'count': 3},
            },
            {'train': 0.5, 'test': 0.5},
            seed=mock.ANY
        )

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_passes_issuer_separated_seed(self, split_mock, splitter):
        splitter.split(
            {'train': 0.5, 'test': 0.5},
            separate_issuers=True
        )

        r = random.Random(INITIAL_SEED)

        split_mock.assert_called_with(
            mock.ANY,
            mock.ANY,
            seed=r.random()
        )

    #
    # SPLIT_BY_AUDIO_DURATION()
    #

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_audio_duration(self, split_mock, splitter):
        split_mock.return_value = {
            'train': ['utt-1', 'utt-3'],
            'test': ['utt-3', 'utt-4'],
            'dev': ['utt-5', 'utt-6'],
        }
        res = splitter.split_by_audio_duration({'train': 0.6, 'test': 0.2, 'dev': 0.2})

        assert res['train'].utterances.keys() == {'utt-1', 'utt-3'}
        assert res['test'].utterances.keys() == {'utt-3', 'utt-4'}
        assert res['dev'].utterances.keys() == {'utt-5', 'utt-6'}

        split_mock.assert_called_with(
            {
                'utt-1': {'duration': 2595},
                'utt-2': {'duration': 2595},
                'utt-3': {'duration': 15000},
                'utt-4': {'duration': 10000},
                'utt-5': {'duration': 15000},
                'utt-6': {'duration': 15000},
                'utt-7': {'duration': 10000},
                'utt-8': {'duration': 15000},
            },
            {'train': 0.6, 'test': 0.2, 'dev': 0.2},
            seed=mock.ANY
        )

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_audio_duration_passes_seed(self, split_mock, splitter):
        splitter.split_by_audio_duration(
            {'train': 0.5, 'test': 0.5},
            separate_issuers=False
        )

        r = random.Random(INITIAL_SEED)

        split_mock.assert_called_with(
            mock.ANY,
            mock.ANY,
            seed=r.random()
        )

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_audio_duration_issuer_separated(self, split_mock, splitter):
        split_mock.return_value = {
            'train': ['spk-1', 'spk-2'],
            'test': ['spk-3']
        }
        res = splitter.split_by_audio_duration(
            {'train': 0.5, 'test': 0.5},
            separate_issuers=True
        )

        assert res['train'].utterances.keys() == {
            'utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'
        }
        assert res['test'].utterances.keys() == {
            'utt-6', 'utt-7', 'utt-8'
        }

        split_mock.assert_called_with(
            {
                'spk-1': {'duration': 5190},
                'spk-2': {'duration': 40000},
                'spk-3': {'duration': 40000},
            },
            {'train': 0.5, 'test': 0.5},
            seed=mock.ANY
        )

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_audio_duration_passes_issuer_separated_seed(self, split_mock, splitter):
        splitter.split_by_audio_duration(
            {'train': 0.5, 'test': 0.5},
            separate_issuers=True
        )

        r = random.Random(INITIAL_SEED)

        split_mock.assert_called_with(
            mock.ANY,
            mock.ANY,
            seed=r.random()
        )

    #
    # SPLIT_BY_LABEL_LENGTH()
    #

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_label_length(self, split_mock, splitter):
        split_mock.return_value = {
            'train': ['utt-1', 'utt-3'],
            'test': ['utt-3', 'utt-4'],
            'dev': ['utt-5', 'utt-6'],
        }
        res = splitter.split_by_label_length({'train': 0.6, 'test': 0.2, 'dev': 0.2})

        assert res['train'].utterances.keys() == {'utt-1', 'utt-3'}
        assert res['test'].utterances.keys() == {'utt-3', 'utt-4'}
        assert res['dev'].utterances.keys() == {'utt-5', 'utt-6'}

        split_mock.assert_called_with(
            {
                'utt-1': {'length': 16},
                'utt-2': {'length': 16},
                'utt-3': {'length': 11},
                'utt-4': {'length': 16},
                'utt-5': {'length': 6},
                'utt-6': {'length': 16},
                'utt-7': {'length': 11},
                'utt-8': {'length': 5},
            },
            {'train': 0.6, 'test': 0.2, 'dev': 0.2},
            seed=mock.ANY
        )

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_label_length_passes_seed(self, split_mock, splitter):
        splitter.split_by_label_length(
            {'train': 0.5, 'test': 0.5},
            separate_issuers=False
        )

        r = random.Random(INITIAL_SEED)

        split_mock.assert_called_with(
            mock.ANY,
            mock.ANY,
            seed=r.random()
        )

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_label_length_only_uses_given_label_list(self, split_mock):
        corpus = resources.create_multi_label_corpus()
        splitter = splitting.Splitter(corpus, random_seed=INITIAL_SEED)

        for utt in corpus.utterances.values():
            utt.set_label_list(annotations.LabelList.create_single(
                'another label', idx='some-idx'
            ))

        split_mock.return_value = {
            'train': ['utt-1', 'utt-3'],
            'test': ['utt-3', 'utt-4'],
            'dev': ['utt-5', 'utt-6'],
        }
        res = splitter.split_by_label_length(
            {'train': 0.6, 'test': 0.2, 'dev': 0.2},
            label_list_idx='default'
        )

        assert res['train'].utterances.keys() == {'utt-1', 'utt-3'}
        assert res['test'].utterances.keys() == {'utt-3', 'utt-4'}
        assert res['dev'].utterances.keys() == {'utt-5', 'utt-6'}

        split_mock.assert_called_with(
            {
                'utt-1': {'length': 16},
                'utt-2': {'length': 16},
                'utt-3': {'length': 11},
                'utt-4': {'length': 16},
                'utt-5': {'length': 6},
                'utt-6': {'length': 16},
                'utt-7': {'length': 11},
                'utt-8': {'length': 5},
            },
            {'train': 0.6, 'test': 0.2, 'dev': 0.2},
            seed=mock.ANY
        )

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_label_length_issuer_separated(self, split_mock, splitter):
        split_mock.return_value = {
            'train': ['spk-1', 'spk-2'],
            'test': ['spk-3']
        }
        res = splitter.split_by_label_length(
            {'train': 0.5, 'test': 0.5},
            separate_issuers=True
        )

        assert res['train'].utterances.keys() == {
            'utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'
        }
        assert res['test'].utterances.keys() == {
            'utt-6', 'utt-7', 'utt-8'
        }

        split_mock.assert_called_with(
            {
                'spk-1': {'length': 32},
                'spk-2': {'length': 33},
                'spk-3': {'length': 32},
            },
            {'train': 0.5, 'test': 0.5},
            seed=mock.ANY
        )

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_label_length_passes_issuer_separated_seed(self, split_mock, splitter):
        splitter.split_by_label_length(
            {'train': 0.5, 'test': 0.5},
            separate_issuers=True
        )

        r = random.Random(INITIAL_SEED)

        split_mock.assert_called_with(
            mock.ANY,
            mock.ANY,
            seed=r.random()
        )

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_label_length_issuer_separated_only_uses_given_label_list(self, split_mock):
        corpus = resources.create_multi_label_corpus()
        splitter = splitting.Splitter(corpus, random_seed=INITIAL_SEED)

        for utt in corpus.utterances.values():
            utt.set_label_list(annotations.LabelList.create_single(
                'another label', idx='some-idx'
            ))

        split_mock.return_value = {
            'train': ['spk-1', 'spk-2'],
            'test': ['spk-3']
        }
        res = splitter.split_by_label_length(
            {'train': 0.5, 'test': 0.5},
            label_list_idx='default',
            separate_issuers=True
        )

        assert res['train'].utterances.keys() == {
            'utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'
        }
        assert res['test'].utterances.keys() == {
            'utt-6', 'utt-7', 'utt-8'
        }

        split_mock.assert_called_with(
            {
                'spk-1': {'length': 32},
                'spk-2': {'length': 33},
                'spk-3': {'length': 32},
            },
            {'train': 0.5, 'test': 0.5},
            seed=mock.ANY
        )

    #
    # SPLIT_BY_LABEL_OCCURENCE()
    #

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_label_occurence(self, split_mock, splitter):
        split_mock.return_value = {
            'train': ['utt-1', 'utt-3'],
            'test': ['utt-3', 'utt-4'],
            'dev': ['utt-5', 'utt-6'],
        }
        res = splitter.split_by_label_occurence({'train': 0.6, 'test': 0.2, 'dev': 0.2})

        assert res['train'].utterances.keys() == {'utt-1', 'utt-3'}
        assert res['test'].utterances.keys() == {'utt-3', 'utt-4'}
        assert res['dev'].utterances.keys() == {'utt-5', 'utt-6'}

        split_mock.assert_called_with(
            {
                'utt-1': {'music': 2, 'speech': 1},
                'utt-2': {'music': 2, 'speech': 1},
                'utt-3': {'music': 1, 'speech': 1},
                'utt-4': {'music': 2, 'speech': 1},
                'utt-5': {'speech': 1},
                'utt-6': {'music': 2, 'speech': 1},
                'utt-7': {'music': 1, 'speech': 1},
                'utt-8': {'music': 1},
            },
            {'train': 0.6, 'test': 0.2, 'dev': 0.2},
            seed=mock.ANY
        )

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_label_occurence_passes_seed(self, split_mock, splitter):
        splitter.split_by_label_occurence(
            {'train': 0.5, 'test': 0.5},
            separate_issuers=False
        )

        r = random.Random(INITIAL_SEED)

        split_mock.assert_called_with(
            mock.ANY,
            mock.ANY,
            seed=r.random()
        )

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_label_occurence_issuer_separated(self, split_mock, splitter):
        split_mock.return_value = {
            'train': ['spk-1', 'spk-2'],
            'test': ['spk-3']
        }
        res = splitter.split_by_label_occurence(
            {'train': 0.5, 'test': 0.5},
            separate_issuers=True
        )

        assert res['train'].utterances.keys() == {
            'utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'
        }
        assert res['test'].utterances.keys() == {
            'utt-6', 'utt-7', 'utt-8'
        }

        split_mock.assert_called_with(
            {
                'spk-1': {'music': 4, 'speech': 2},
                'spk-2': {'music': 3, 'speech': 3},
                'spk-3': {'music': 4, 'speech': 2},
            },
            {'train': 0.5, 'test': 0.5},
            seed=mock.ANY
        )

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_label_occurence_passes_issuer_separated_seed(self, split_mock, splitter):
        splitter.split_by_label_occurence(
            {'train': 0.5, 'test': 0.5},
            separate_issuers=True
        )

        r = random.Random(INITIAL_SEED)

        split_mock.assert_called_with(
            mock.ANY,
            mock.ANY,
            seed=r.random()
        )

    #
    # SPLIT_BY_LABEL_DURATION()
    #

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_label_duration(self, split_mock, splitter):
        split_mock.return_value = {
            'train': ['utt-1', 'utt-3'],
            'test': ['utt-3', 'utt-4'],
            'dev': ['utt-5', 'utt-6'],
        }
        res = splitter.split_by_label_duration({'train': 0.6, 'test': 0.2, 'dev': 0.2})

        assert res['train'].utterances.keys() == {'utt-1', 'utt-3'}
        assert res['test'].utterances.keys() == {'utt-3', 'utt-4'}
        assert res['dev'].utterances.keys() == {'utt-5', 'utt-6'}

        split_mock.assert_called_with(
            {
                'utt-1': {'music': 7, 'speech': 7},
                'utt-2': {'music': 7, 'speech': 7},
                'utt-3': {'music': 1, 'speech': 4},
                'utt-4': {'music': 7, 'speech': 7},
                'utt-5': {'speech': 7},
                'utt-6': {'music': 7, 'speech': 7},
                'utt-7': {'music': 5, 'speech': 6},
                'utt-8': {'music': 10},
            },
            {'train': 0.6, 'test': 0.2, 'dev': 0.2},
            seed=mock.ANY
        )

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_label_duration_passes_seed(self, split_mock, splitter):
        splitter.split_by_label_duration(
            {'train': 0.5, 'test': 0.5},
            separate_issuers=False
        )

        r = random.Random(INITIAL_SEED)

        split_mock.assert_called_with(
            mock.ANY,
            mock.ANY,
            seed=r.random()
        )

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_label_duration_issuer_separated(self, split_mock, splitter):
        split_mock.return_value = {
            'train': ['spk-1', 'spk-2'],
            'test': ['spk-3']
        }
        res = splitter.split_by_label_duration(
            {'train': 0.5, 'test': 0.5},
            separate_issuers=True
        )

        assert res['train'].utterances.keys() == {
            'utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'
        }
        assert res['test'].utterances.keys() == {
            'utt-6', 'utt-7', 'utt-8'
        }

        split_mock.assert_called_with(
            {
                'spk-1': {'music': 14, 'speech': 14},
                'spk-2': {'music': 8, 'speech': 18},
                'spk-3': {'music': 22, 'speech': 13},
            },
            {'train': 0.5, 'test': 0.5},
            seed=mock.ANY
        )

    @mock.patch('audiomate.corpus.subset.utils.get_identifiers_splitted_by_weights')
    def test_split_by_label_duration_passes_issuer_separated_seed(self, split_mock, splitter):
        splitter.split_by_label_duration(
            {'train': 0.5, 'test': 0.5},
            separate_issuers=True
        )

        r = random.Random(INITIAL_SEED)

        split_mock.assert_called_with(
            mock.ANY,
            mock.ANY,
            seed=r.random()
        )



import copy
import os
import shutil

from audiomate import tracks
from audiomate import containers
from audiomate import issuers
from audiomate.utils import naming
from audiomate.utils import audio
from . import base
from . import subset

DEFAULT_FILE_SUBDIR = 'files'
DEFAULT_FEAT_SUBDIR = 'features'


class Corpus(base.CorpusView):
    """
    The Corpus class represents a single corpus.
    It extends :py:class:`audiomate.corpus.CorpusView` with the functionality for loading and saving.
    Furthermore it provides the functionality for adding/modifying assets of the corpus like tracks
    and utterances.

    Args:
        path (str): Path where the corpus is stored. (Optional)
    """

    def __init__(self, path=None):
        super(Corpus, self).__init__()

        self.path = path
        self._tracks = {}
        self._utterances = {}
        self._issuers = {}
        self._feature_containers = {}
        self._subviews = {}

    @property
    def name(self):
        if self.path is None:
            return 'undefined'
        else:
            return os.path.basename(os.path.abspath(self.path))

    @property
    def tracks(self):
        return self._tracks

    @property
    def utterances(self):
        return self._utterances

    @property
    def issuers(self):
        return self._issuers

    @property
    def feature_containers(self):
        return self._feature_containers

    @property
    def subviews(self):
        return self._subviews

    #
    #   IO
    #

    def save(self, writer=None):
        """
        If self.path is defined, it tries to save the corpus at the given path.
        """

        if self.path is None:
            raise ValueError('No path given to save the data set.')

        self.save_at(self.path, writer)

    def save_at(self, path, writer=None):
        """
        Save this corpus at the given path. If the path differs from the current path set, the path
        gets updated.

        Parameters:
            path (str): Path to save the data set to.
            writer (str, CorpusWriter): The writer or the name of the reader to use.
        """

        if writer is None:
            from . import io
            writer = io.DefaultWriter()
        elif type(writer) == str:
            # If a loader is given as string, try to create such a loader.
            from . import io
            writer = io.create_writer_of_type(writer)

        writer.save(self, path)

        self.path = path

    @classmethod
    def load(cls, path, reader=None):
        """
        Loads the corpus from the given path, using the given reader. If no reader is given the
        :py:class:`audiomate.corpus.io.DefaultReader` is used.

        Args:
            path (str): Path to load the corpus from.
            reader (str, CorpusReader): The reader or the name of the reader to use.

        Returns:
            Corpus: The loaded corpus.
        """

        if reader is None:
            from . import io
            reader = io.DefaultReader()

        elif type(reader) == str:
            from . import io
            reader = io.create_reader_of_type(reader)

        return reader.load(path)

    #
    # Track
    #

    def new_file(self, path, track_idx, copy_file=False):
        """
        Adds a new audio file to the corpus with the given data.

        Parameters:
            path (str): Path of the file to add.
            track_idx (str): The id to associate the file-track with.
            copy_file (bool): If True the file is copied to the data set folder, otherwise the given
                              path is used directly.

        Returns:
            FileTrack: The newly added file.
        """

        new_file_idx = track_idx
        new_file_path = os.path.abspath(path)

        # Add index to idx if already existing
        if new_file_idx in self._tracks.keys():
            new_file_idx = naming.index_name_if_in_list(new_file_idx, self._tracks.keys())

        # Copy file to default file dir
        if copy_file:
            if not os.path.isdir(self.path):
                raise ValueError('To copy file the dataset needs to have a path.')

            __, ext = os.path.splitext(path)

            new_file_folder = os.path.join(self.path, DEFAULT_FILE_SUBDIR)
            new_file_path = os.path.join(new_file_folder, '{}{}'.format(new_file_idx, ext))
            os.makedirs(new_file_folder, exist_ok=True)
            shutil.copy(path, new_file_path)

        # Create file obj
        new_file = tracks.FileTrack(new_file_idx, new_file_path)
        self._tracks[new_file_idx] = new_file

        return new_file

    def import_tracks(self, import_tracks):
        """
        Add the given tracks/track to the corpus.
        If any of the given track-ids already exists, a suffix is appended so it is unique.

        Args:
            import_tracks (list): Either a list of or a single :py:class:`audiomate.tracks.Track`.

        Returns:
            dict: A dictionary containing track-idx mappings (old-track-idx/track-instance).
                  If a track is imported, whose idx already exists this mapping can be used to check
                  the new id.
        """

        if isinstance(import_tracks, tracks.Track):
            import_tracks = [import_tracks]

        idx_mapping = {}

        for track in import_tracks:
            idx_mapping[track.idx] = track

            # Add index to idx if already existing
            if track.idx in self._tracks.keys():
                track.idx = naming.index_name_if_in_list(track.idx, self._tracks.keys())

            self._tracks[track.idx] = track

        return idx_mapping

    #
    #   Utterances
    #

    def new_utterance(self, utterance_idx, track_idx, issuer_idx=None, start=0, end=float('inf')):
        """
        Add a new utterance to the corpus with the given data.

        Parameters:
            track_idx (str): The track id the utterance is in.
            utterance_idx (str): The id to associate with the utterance.
                                 If None or already exists, one is generated.
            issuer_idx (str): The issuer id to associate with the utterance.
            start (float): Start of the utterance within the track [seconds].
            end (float): End of the utterance within the track [seconds].
                         ``inf`` equals the end of the track.

        Returns:
            Utterance: The newly added utterance.
        """

        new_utt_idx = utterance_idx

        # Check if there is a track with the given idx
        if track_idx not in self._tracks.keys():
            raise ValueError('Track with id {} does not exist!'.format(track_idx))

        # Check if issuer exists
        issuer = None

        if issuer_idx is not None:
            if issuer_idx not in self._issuers.keys():
                raise ValueError('Issuer with id {} does not exist!'.format(issuer_idx))
            else:
                issuer = self._issuers[issuer_idx]

        # Add index to idx if already existing
        if new_utt_idx in self._utterances.keys():
            new_utt_idx = naming.index_name_if_in_list(new_utt_idx, self._utterances.keys())

        new_utt = tracks.Utterance(new_utt_idx,
                                   self.tracks[track_idx],
                                   issuer=issuer,
                                   start=start,
                                   end=end)

        self._utterances[new_utt_idx] = new_utt

        return new_utt

    def import_utterances(self, utterances):
        """
        Add the given utterances/utterance to the corpus.
        If any of the given utterance-ids already exists, a suffix is appended so it is unique.

        Args:
            utterances (list): Either a list of or a single :py:class:`audiomate.tracks.Utterance`.

        Returns:
            dict: A dictionary containing idx mappings (old-utterance-idx/utterance-instance).
                  If a utterance is imported, whose id already exists this mapping can be used to
                  check the new id.
        """

        if isinstance(utterances, tracks.Utterance):
            utterances = [utterances]

        idx_mapping = {}

        for utterance in utterances:
            idx_mapping[utterance.idx] = utterance

            # Check if there is a track with the given idx
            if not self.contains_track(utterance.track):
                raise ValueError('Track with id {} is not in the corpus.'.format(utterance.track.idx, utterance.idx))

            # Check if there is a issuer with the given idx
            if utterance.issuer is not None and not self.contains_issuer(utterance.issuer):
                raise ValueError('No issuer in corpus with id {} to add utterance {}.'.format(
                    utterance.issuer.idx, utterance.idx))

            # Add index to idx if already existing
            if utterance.idx in self._utterances.keys():
                utterance.idx = naming.index_name_if_in_list(utterance.idx, self._utterances.keys())

            self._utterances[utterance.idx] = utterance

        return idx_mapping

    #
    #   Issuer
    #

    def new_issuer(self, issuer_idx, info=None):
        """
        Add a new issuer to the dataset with the given data.

        Parameters:
            issuer_idx (str): The id to associate the issuer with. If None or already exists, one is
                              generated.
            info (dict, list): Additional info of the issuer.

        Returns:
            Issuer: The newly added issuer.
        """

        new_issuer_idx = issuer_idx

        # Add index to idx if already existing
        if new_issuer_idx in self._issuers.keys():
            new_issuer_idx = naming.index_name_if_in_list(new_issuer_idx, self._issuers.keys())

        new_issuer = issuers.Issuer(new_issuer_idx, info=info)
        self._issuers[new_issuer_idx] = new_issuer

        return new_issuer

    def import_issuers(self, new_issuers):
        """
        Add the given issuers/issuer to the corpus.
        If any of the given issuer-ids already exists, a suffix is appended so it is unique.

        Args:
            issuers (list): Either a list of or a single :py:class:`audiomate.issuers.Issuer`.

        Returns:
            dict: A dictionary containing idx mappings (old-issuer-idx/issuer-instance).
                  If a issuer is imported, whose id already exists this mapping can be used to check
                  the new id.
        """

        if isinstance(new_issuers, issuers.Issuer):
            new_issuers = [new_issuers]

        idx_mapping = {}

        for issuer in new_issuers:
            idx_mapping[issuer.idx] = issuer

            # Add index to idx if already existing
            if issuer.idx in self._issuers.keys():
                issuer.idx = naming.index_name_if_in_list(issuer.idx, self._issuers.keys())

            self._issuers[issuer.idx] = issuer

        return idx_mapping

    #
    #   FEATURES
    #

    def new_feature_container(self, idx, path=None):
        """
        Add a new feature container with the given data.

        Parameters:
            idx (str): An unique identifier within the dataset.
            path (str): The path to store the feature file. If None a default path is used.

        Returns:
            FeatureContainer: The newly added feature-container.
        """

        new_feature_idx = idx
        new_feature_path = path

        # Add index to idx if already existing
        if new_feature_idx in self._feature_containers.keys():
            new_feature_idx = naming.index_name_if_in_list(new_feature_idx,
                                                           self._feature_containers.keys())

        # Set default path if none given
        if new_feature_path is None:
            if not os.path.isdir(self.path):
                raise ValueError('To copy file the dataset needs to have a path.')

            new_feature_path = os.path.join(self.path, DEFAULT_FEAT_SUBDIR, new_feature_idx)
        else:
            new_feature_path = os.path.abspath(new_feature_path)

        feat_container = containers.FeatureContainer(new_feature_path)
        self._feature_containers[new_feature_idx] = feat_container

        return feat_container

    #
    #   Subviews
    #

    def import_subview(self, idx, subview):
        """
        Add the given subview to the corpus.

        Args:
            idx (str): An idx that is unique in the corpus for identifying the subview.
                       If already a subview exists with the given id it will be overridden.
            subview (Subview): The subview to add.
        """

        subview.corpus = self
        self._subviews[idx] = subview

    #
    #   Merge
    #

    def merge_corpus(self, corpus):
        """
        Merge the given corpus into this corpus. All assets (tracks, utterances, issuers, ...) are copied into
        this corpus. If any ids (utt-idx, track-idx, issuer-idx, subview-idx, ...) are occurring in both corpora,
        the ids from the merging corpus are suffixed by a number (starting from 1 until no other is matching).

        Args:
            corpus (CorpusView): The corpus to merge.
        """

        # Create a copy, so objects aren't changed in the original merging corpus
        merging_corpus = Corpus.from_corpus(corpus)

        self.import_tracks(corpus.tracks.values())
        self.import_issuers(corpus.issuers.values())
        utterance_idx_mapping = self.import_utterances(corpus.utterances.values())

        for subview_idx, subview in merging_corpus.subviews.items():
            for filter in subview.filter_criteria:
                if isinstance(filter, subset.MatchingUtteranceIdxFilter):
                    new_filtered_utt_ids = set()
                    for utt_idx in filter.utterance_idxs:
                        new_filtered_utt_ids.add(utterance_idx_mapping[utt_idx].idx)
                    filter.utterance_idxs = new_filtered_utt_ids

            new_idx = naming.index_name_if_in_list(subview_idx, self.subviews.keys())
            self.import_subview(new_idx, subview)

        for feat_container_idx, feat_container in merging_corpus.feature_containers.items():
            self.new_feature_container(feat_container_idx, feat_container.path)

    #
    #   VARIA
    #

    def relocate_audio_to_single_container(self, target_path):
        """
        Copies every track to a single container.
        Afterwards all tracks in the container are linked against
        this single container.
        """

        cont = containers.AudioContainer(target_path)
        cont.open()

        new_tracks = {}

        # First create a new container track for all existing tracks
        for track in self.tracks.values():
            sr = track.sampling_rate
            samples = track.read_samples()

            cont.set(track.idx, samples, sr)
            new_track = tracks.ContainerTrack(track.idx, cont)

            new_tracks[track.idx] = new_track

        # Update track list of corpus
        self._tracks = new_tracks

        # Update utterances to point to new tracks
        for utterance in self.utterances.values():
            new_track = self.tracks[utterance.track.idx]
            utterance.track = new_track

        cont.close()

    def relocate_audio_to_wav_files(self, target_path):
        """
        Copies every track to its own wav file in the given folder.
        Every track will be stored at ``target_path/track_id.wav``.
        """

        if not os.path.isdir(target_path):
            os.makedirs(target_path)

        new_tracks = {}

        # First create a new container track for all existing tracks
        for track in self.tracks.values():
            track_path = os.path.join(target_path, '{}.wav'.format(track.idx))
            sr = track.sampling_rate
            samples = track.read_samples()

            audio.write_wav(track_path, samples, sr=sr)
            new_track = tracks.FileTrack(track.idx, track_path)

            new_tracks[track.idx] = new_track

        # Update track list of corpus
        self._tracks = new_tracks

        # Update utterances to point to new tracks
        for utterance in self.utterances.values():
            new_track = self.tracks[utterance.track.idx]
            utterance.track = new_track

    #
    #   Creation
    #

    @classmethod
    def from_corpus(cls, corpus):
        """
        Create a new modifiable corpus from any other CorpusView.
        This for example can be used to create a independent modifiable corpus from a subview.

        Args:
            corpus (CorpusView): The corpus to create a copy from.

        Returns:
            Corpus: A new corpus with the same data as the given one.
        """

        ds = Corpus()

        # Tracks
        tracks = copy.deepcopy(list(corpus.tracks.values()))
        track_mapping = ds.import_tracks(tracks)

        # Issuers
        issuers = copy.deepcopy(list(corpus.issuers.values()))
        issuer_mapping = ds.import_issuers(issuers)

        # Utterances, with replacing changed track- and issuer-ids
        utterances = copy.deepcopy(list(corpus.utterances.values()))
        for utterance in utterances:
            utterance.track = track_mapping[utterance.track.idx]

            if utterance.issuer is not None:
                utterance.issuer = issuer_mapping[utterance.issuer.idx]

        ds.import_utterances(utterances)

        # Subviews
        subviews = copy.deepcopy(corpus.subviews)
        for subview_idx, subview in subviews.items():
            ds.import_subview(subview_idx, subview)

        # Feat-Containers
        for feat_container_idx, feature_container in corpus.feature_containers.items():
            ds.new_feature_container(feat_container_idx, feature_container.path)

        return ds

    @classmethod
    def merge_corpora(cls, corpora):
        """
        Merge a list of corpora into one.

        Args:
            corpora (Iterable): An iterable of :py:class:`audiomate.corpus.CorpusView`.

        Returns:
            Corpus: A corpus with the data from all given corpora merged into one.
        """

        ds = Corpus()

        for merging_corpus in corpora:
            ds.merge_corpus(merging_corpus)

        return ds



import os

from audiomate import corpus
from audiomate import issuers
from audiomate.corpus import io

from tests import resources
from . import reader_test as rt


class TestSpeechCommandsReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('speech_commands')
    FILE_TRACK_BASE_PATH = SAMPLE_PATH

    EXPECTED_NUMBER_OF_TRACKS = 13
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('0b77ee66_nohash_0_bed', os.path.join('bed', '0b77ee66_nohash_0.wav')),
        rt.ExpFileTrack('0b77ee66_nohash_1_bed', os.path.join('bed', '0b77ee66_nohash_1.wav')),
        rt.ExpFileTrack('0b77ee66_nohash_2_bed', os.path.join('bed', '0b77ee66_nohash_2.wav')),
        rt.ExpFileTrack('0bde966a_nohash_0_bed', os.path.join('bed', '0bde966a_nohash_0.wav')),
        rt.ExpFileTrack('0bde966a_nohash_1_bed', os.path.join('bed', '0bde966a_nohash_1.wav')),
        rt.ExpFileTrack('0c40e715_nohash_0_bed', os.path.join('bed', '0c40e715_nohash_0.wav')),
        rt.ExpFileTrack('d5c41d6a_nohash_0_marvin', os.path.join('marvin', 'd5c41d6a_nohash_0.wav')),
        rt.ExpFileTrack('d7a58714_nohash_0_marvin', os.path.join('marvin', 'd7a58714_nohash_0.wav')),
        rt.ExpFileTrack('d8a5ace5_nohash_0_marvin', os.path.join('marvin', 'd8a5ace5_nohash_0.wav')),
        rt.ExpFileTrack('0a7c2a8d_nohash_0_one', os.path.join('one', '0a7c2a8d_nohash_0.wav')),
        rt.ExpFileTrack('0b77ee66_nohash_0_one', os.path.join('one', '0b77ee66_nohash_0.wav')),
        rt.ExpFileTrack('c1b7c224_nohash_0_one', os.path.join('one', 'c1b7c224_nohash_0.wav')),
        rt.ExpFileTrack('c1b7c224_nohash_1_one', os.path.join('one', 'c1b7c224_nohash_1.wav')),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 8
    EXPECTED_ISSUERS = [
        rt.ExpSpeaker('0b77ee66', 4, issuers.Gender.UNKNOWN, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('0bde966a', 2, issuers.Gender.UNKNOWN, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('0c40e715', 1, issuers.Gender.UNKNOWN, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('d5c41d6a', 1, issuers.Gender.UNKNOWN, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('d7a58714', 1, issuers.Gender.UNKNOWN, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('d8a5ace5', 1, issuers.Gender.UNKNOWN, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('0a7c2a8d', 1, issuers.Gender.UNKNOWN, issuers.AgeGroup.UNKNOWN, None),
        rt.ExpSpeaker('c1b7c224', 2, issuers.Gender.UNKNOWN, issuers.AgeGroup.UNKNOWN, None),
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 13
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('0b77ee66_nohash_0_bed', '0b77ee66_nohash_0_bed', '0b77ee66', 0, float('inf')),
        rt.ExpUtterance('0b77ee66_nohash_1_bed', '0b77ee66_nohash_1_bed', '0b77ee66', 0, float('inf')),
        rt.ExpUtterance('0b77ee66_nohash_2_bed', '0b77ee66_nohash_2_bed', '0b77ee66', 0, float('inf')),
        rt.ExpUtterance('0bde966a_nohash_0_bed', '0bde966a_nohash_0_bed', '0bde966a', 0, float('inf')),
        rt.ExpUtterance('0bde966a_nohash_1_bed', '0bde966a_nohash_1_bed', '0bde966a', 0, float('inf')),
        rt.ExpUtterance('0c40e715_nohash_0_bed', '0c40e715_nohash_0_bed', '0c40e715', 0, float('inf')),
        rt.ExpUtterance('d5c41d6a_nohash_0_marvin', 'd5c41d6a_nohash_0_marvin', 'd5c41d6a', 0, float('inf')),
        rt.ExpUtterance('d7a58714_nohash_0_marvin', 'd7a58714_nohash_0_marvin', 'd7a58714', 0, float('inf')),
        rt.ExpUtterance('d8a5ace5_nohash_0_marvin', 'd8a5ace5_nohash_0_marvin', 'd8a5ace5', 0, float('inf')),
        rt.ExpUtterance('0a7c2a8d_nohash_0_one', '0a7c2a8d_nohash_0_one', '0a7c2a8d', 0, float('inf')),
        rt.ExpUtterance('0b77ee66_nohash_0_one', '0b77ee66_nohash_0_one', '0b77ee66', 0, float('inf')),
        rt.ExpUtterance('c1b7c224_nohash_0_one', 'c1b7c224_nohash_0_one', 'c1b7c224', 0, float('inf')),
        rt.ExpUtterance('c1b7c224_nohash_1_one', 'c1b7c224_nohash_1_one', 'c1b7c224', 0, float('inf')),
    ]

    EXPECTED_LABEL_LISTS = {
        '0b77ee66_nohash_0_bed': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)],
        '0b77ee66_nohash_1_bed': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)],
        '0b77ee66_nohash_2_bed': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)],
        '0bde966a_nohash_0_bed': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)],
        '0bde966a_nohash_1_bed': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)],
        '0c40e715_nohash_0_bed': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)],
        'd5c41d6a_nohash_0_marvin': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)],
        'd7a58714_nohash_0_marvin': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)],
        'd8a5ace5_nohash_0_marvin': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)],
        '0a7c2a8d_nohash_0_one': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)],
        '0b77ee66_nohash_0_one': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)],
        'c1b7c224_nohash_0_one': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)],
        'c1b7c224_nohash_1_one': [rt.ExpLabelList(corpus.LL_WORD_TRANSCRIPT, 1)],
    }

    EXPECTED_LABELS = {
        '0b77ee66_nohash_0_bed': [rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'bed', 0, float('inf'))],
        '0b77ee66_nohash_1_bed': [rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'bed', 0, float('inf'))],
        '0b77ee66_nohash_2_bed': [rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'bed', 0, float('inf'))],
        '0bde966a_nohash_0_bed': [rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'bed', 0, float('inf'))],
        '0bde966a_nohash_1_bed': [rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'bed', 0, float('inf'))],
        '0c40e715_nohash_0_bed': [rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'bed', 0, float('inf'))],
        'd5c41d6a_nohash_0_marvin': [rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'marvin', 0, float('inf'))],
        'd7a58714_nohash_0_marvin': [rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'marvin', 0, float('inf'))],
        'd8a5ace5_nohash_0_marvin': [rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'marvin', 0, float('inf'))],
        '0a7c2a8d_nohash_0_one': [rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'one', 0, float('inf'))],
        '0b77ee66_nohash_0_one': [rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'one', 0, float('inf'))],
        'c1b7c224_nohash_0_one': [rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'one', 0, float('inf'))],
        'c1b7c224_nohash_1_one': [rt.ExpLabel(corpus.LL_WORD_TRANSCRIPT, 'one', 0, float('inf'))],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 3
    EXPECTED_SUBVIEWS = [
        rt.ExpSubview('train', [
            '0b77ee66_nohash_0_bed',
            '0b77ee66_nohash_1_bed',
            '0b77ee66_nohash_2_bed',
            'd5c41d6a_nohash_0_marvin',
            'c1b7c224_nohash_0_one',
            'c1b7c224_nohash_1_one',
        ]),
        rt.ExpSubview('dev', [
            '0c40e715_nohash_0_bed',
            'd8a5ace5_nohash_0_marvin',
            '0a7c2a8d_nohash_0_one',
        ]),
        rt.ExpSubview('test', [
            '0bde966a_nohash_0_bed',
            '0bde966a_nohash_1_bed',
            'd7a58714_nohash_0_marvin',
            '0b77ee66_nohash_0_one',
        ]),
    ]

    def load(self):
        return io.SpeechCommandsReader().load(self.SAMPLE_PATH)



import os

import numpy as np
import librosa
import h5py

import pytest

from audiomate import tracks
from audiomate import containers
from audiomate import processing

from tests import resources


class ProcessorDummy(processing.Processor):

    def __init__(self, frame_size_scale=1.0, hop_size_scale=1.0):
        self.called_with_data = []
        self.called_with_sr = []
        self.called_with_offset = []
        self.called_with_last = []
        self.called_with_utterance = []
        self.called_with_corpus = []

        self.mock_frame_size_scale = frame_size_scale
        self.mock_hop_size_scale = hop_size_scale

    def process_frames(self, data, sampling_rate, offset=0, last=False, utterance=None, corpus=None):
        self.called_with_data.append(data)
        self.called_with_sr.append(sampling_rate)
        self.called_with_offset.append(offset)
        self.called_with_last.append(last)
        self.called_with_utterance.append(utterance)
        self.called_with_corpus.append(corpus)

        return data

    def frame_transform(self, frame_size, hop_size):
        tf_frame_size = self.mock_frame_size_scale * frame_size
        tf_hop_size = self.mock_hop_size_scale * hop_size

        return tf_frame_size, tf_hop_size


@pytest.fixture()
def processor():
    return ProcessorDummy()


@pytest.fixture()
def sample_utterance():
    file_track = tracks.FileTrack('test_file', resources.sample_wav_file('wav_1.wav'))
    utterance = tracks.Utterance('test', file_track)
    return utterance


class TestProcessor:

    #
    # process_track
    #

    def test_process_track(self, processor, tmpdir):
        wav_path = os.path.join(tmpdir.strpath, 'file.wav')
        wav_content = np.random.random(22)

        librosa.output.write_wav(wav_path, wav_content, 4)
        file_track = tracks.FileTrack('idx', wav_path)

        processed = processor.process_track(file_track, frame_size=4, hop_size=2)

        assert processed.shape == (10, 4)
        assert processed.dtype == np.float32
        assert np.allclose(processed[0], wav_content[0:4], atol=0.0001)
        assert np.allclose(processed[9], wav_content[18:22], atol=0.0001)

        assert processor.called_with_sr == [4]
        assert processor.called_with_offset == [0]
        assert processor.called_with_last == [True]
        assert processor.called_with_utterance == [None]
        assert processor.called_with_corpus == [None]

    def test_process_track_smaller_than_frame_size(self, processor, tmpdir):
        wav_path = os.path.join(tmpdir.strpath, 'file.wav')
        wav_content = np.random.random(22)

        librosa.output.write_wav(wav_path, wav_content, 16000)
        file_track = tracks.FileTrack('idx', wav_path)

        processed = processor.process_track(file_track, frame_size=4096, hop_size=2048, sr=16000)

        assert processed.shape == (1, 4096)
        assert np.allclose(processed[0], np.pad(wav_content, (0, 4074), mode='constant'), atol=0.0001)

        assert processor.called_with_sr == [16000]
        assert processor.called_with_offset == [0]
        assert processor.called_with_last == [True]
        assert processor.called_with_utterance == [None]
        assert processor.called_with_corpus == [None]

    def test_process_empty_track_raises_error(self, processor, tmpdir):
        wav_path = os.path.join(tmpdir.strpath, 'file.wav')
        wav_content = np.random.random(0)

        librosa.output.write_wav(wav_path, wav_content, 16000)
        file_track = tracks.FileTrack('idx', wav_path)

        with pytest.raises(ValueError):
            processor.process_track(file_track, frame_size=4096, hop_size=2048, sr=16000)

    def test_process_track_with_downsampling(self, processor, tmpdir):
        wav_path = os.path.join(tmpdir.strpath, 'file.wav')
        wav_content = np.random.random(22)

        librosa.output.write_wav(wav_path, wav_content, 4)
        file_track = tracks.FileTrack('idx', wav_path)

        processed = processor.process_track(file_track, frame_size=4, hop_size=2, sr=2)

        assert processed.shape == (5, 4)

        assert processor.called_with_sr == [2]
        assert processor.called_with_offset == [0]
        assert processor.called_with_last == [True]
        assert processor.called_with_utterance == [None]
        assert processor.called_with_corpus == [None]

    #
    #   process_track_online
    #

    def test_process_track_online(self, processor, tmpdir):
        wav_path = os.path.join(tmpdir.strpath, 'file.wav')
        wav_content = np.random.random(174)

        librosa.output.write_wav(wav_path, wav_content, 16000)
        track = tracks.FileTrack('idx', wav_path)

        chunks = list(processor.process_track_online(track, frame_size=20, hop_size=10, chunk_size=8))

        assert len(chunks) == 3
        assert np.allclose(chunks[0][0], wav_content[0:20], atol=0.0001)
        assert np.allclose(chunks[2][-1], np.pad(wav_content[160:], (0, 6), mode='constant'), atol=0.0001)
        assert chunks[0].dtype == np.float32

        assert processor.called_with_sr == [16000, 16000, 16000]
        assert processor.called_with_offset == [0, 8, 16]
        assert processor.called_with_last == [False, False, True]
        assert processor.called_with_utterance == [None, None, None]
        assert processor.called_with_corpus == [None, None, None]

    def test_process_track_online_no_rest_frame(self, processor, tmpdir):
        wav_path = os.path.join(tmpdir.strpath, 'file.wav')
        wav_content = np.random.random(170)

        librosa.output.write_wav(wav_path, wav_content, 16000)
        track = tracks.FileTrack('idx', wav_path)

        chunks = list(processor.process_track_online(track, frame_size=20, hop_size=10, chunk_size=8))

        assert len(chunks) == 2
        assert np.allclose(chunks[0][0], wav_content[0:20], atol=0.0001)
        assert np.allclose(chunks[1][-1], wav_content[150:], atol=0.0001)

        assert processor.called_with_sr == [16000, 16000]
        assert processor.called_with_offset == [0, 8]
        assert processor.called_with_last == [False, True]
        assert processor.called_with_utterance == [None, None]
        assert processor.called_with_corpus == [None, None]

    #
    #   process_utterance_...
    #

    def test_process_utterance(self, processor, sample_utterance):
        data = processor.process_utterance(sample_utterance, frame_size=4096, hop_size=2048)

        assert data.shape == (20, 4096)

    def test_process_utterance_with_start_end(self, processor, sample_utterance):
        sample_utterance.start = 1.0
        sample_utterance.end = 1.5

        data = processor.process_utterance(sample_utterance, frame_size=4096, hop_size=2048)

        assert data.shape == (3, 4096)

    def test_process_utterance_online(self, processor, sample_utterance):
        chunks = list(processor.process_utterance_online(sample_utterance, frame_size=4096,
                                                         hop_size=2048, chunk_size=4))

        assert len(chunks) == 5
        assert np.vstack(chunks).shape == (20, 4096)

    def test_process_utterance_online_with_start_end(self, processor, sample_utterance):
        sample_utterance.start = 1.0
        sample_utterance.end = 1.5

        chunks = list(processor.process_utterance_online(sample_utterance, frame_size=4096,
                                                         hop_size=2048, chunk_size=4))

        assert len(chunks) == 1
        assert np.vstack(chunks).shape == (3, 4096)

    #
    #   process_corpus
    #

    def test_process_corpus(self, processor, tmpdir):
        ds = resources.create_dataset()
        feat_path = os.path.join(tmpdir.strpath, 'feats')

        processor.process_corpus(ds, feat_path, frame_size=4096, hop_size=2048)

        with h5py.File(feat_path, 'r') as f:
            utts = set(f.keys())

            assert utts == set(ds.utterances.keys())

            assert f['utt-1'].shape == (20, 4096)
            assert f['utt-2'].shape == (20, 4096)
            assert f['utt-3'].shape == (11, 4096)
            assert f['utt-4'].shape == (7, 4096)
            assert f['utt-5'].shape == (20, 4096)

    def test_process_corpus_with_downsampling(self, processor, tmpdir):
        ds = resources.create_dataset()
        feat_path = os.path.join(tmpdir.strpath, 'feats')

        processor.process_corpus(ds, feat_path, frame_size=4096, hop_size=2048, sr=8000)

        with h5py.File(feat_path, 'r') as f:
            utts = set(f.keys())

            assert utts == set(ds.utterances.keys())

            assert f['utt-1'].shape == (10, 4096)
            assert f['utt-2'].shape == (10, 4096)
            assert f['utt-3'].shape == (5, 4096)
            assert f['utt-4'].shape == (3, 4096)
            assert f['utt-5'].shape == (10, 4096)

    def test_process_corpus_sets_container_attributes(self, processor, tmpdir):
        ds = resources.create_dataset()
        feat_path = os.path.join(tmpdir.strpath, 'feats')

        feat_container = processor.process_corpus(ds, feat_path, frame_size=4096, hop_size=2048)

        with feat_container:
            assert feat_container.frame_size == 4096
            assert feat_container.hop_size == 2048
            assert feat_container.sampling_rate == 16000

    def test_process_corpus_sets_container_attributes_with_downsampling(self, processor, tmpdir):
        ds = resources.create_dataset()
        feat_path = os.path.join(tmpdir.strpath, 'feats')

        feat_container = processor.process_corpus(ds, feat_path, frame_size=4096, hop_size=2048, sr=8000)

        with feat_container:
            assert feat_container.frame_size == 4096
            assert feat_container.hop_size == 2048
            assert feat_container.sampling_rate == 8000

    def test_process_corpus_with_frame_hop_size_change_stores_correct(self, processor, tmpdir):
        ds = resources.create_dataset()
        feat_path = os.path.join(tmpdir.strpath, 'feats')

        processor.mock_frame_size_scale = 2.5
        processor.mock_hop_size_scale = 5
        processor.process_corpus(ds, feat_path, frame_size=4096, hop_size=2048)

        fc = containers.FeatureContainer(feat_path)
        fc.open()

        assert fc.frame_size == 10240
        assert fc.hop_size == 10240

        fc.close()

    #
    #   process_corpus_online
    #

    def test_process_corpus_online(self, processor, tmpdir):
        ds = resources.create_dataset()
        feat_path = os.path.join(tmpdir.strpath, 'feats')

        processor.process_corpus_online(ds, feat_path, frame_size=4096, hop_size=2048)

        with h5py.File(feat_path, 'r') as f:
            utts = set(f.keys())

            assert utts == set(ds.utterances.keys())

            assert f['utt-1'].shape == (20, 4096)
            assert f['utt-2'].shape == (20, 4096)
            assert f['utt-3'].shape == (11, 4096)
            assert f['utt-4'].shape == (7, 4096)
            assert f['utt-5'].shape == (20, 4096)

    def test_process_corpus_online_sets_container_attributes(self, processor, tmpdir):
        ds = resources.create_dataset()
        feat_path = os.path.join(tmpdir.strpath, 'feats')

        feat_container = processor.process_corpus_online(ds, feat_path, frame_size=4096, hop_size=2048)

        with feat_container:
            assert feat_container.frame_size == 4096
            assert feat_container.hop_size == 2048
            assert feat_container.sampling_rate == 16000

    def test_process_corpus_online_ignore_returning_none(self, processor, tmpdir):
        ds = resources.create_dataset()
        feat_path = os.path.join(tmpdir.strpath, 'feats')

        def return_none(*args, **kwargs):
            return None

        processor.process_frames = return_none
        processor.process_corpus_online(ds, feat_path, frame_size=4096, hop_size=2048)

        assert True

    def test_process_corpus_online_with_frame_hop_size_change_stores_correct(self, processor, tmpdir):
        ds = resources.create_dataset()
        feat_path = os.path.join(tmpdir.strpath, 'feats')

        processor.mock_frame_size_scale = 0.5
        processor.mock_hop_size_scale = 0.25
        processor.process_corpus_online(ds, feat_path, frame_size=4096, hop_size=2048)

        fc = containers.FeatureContainer(feat_path)
        fc.open()

        assert fc.frame_size == 2048
        assert fc.hop_size == 512

        fc.close()

    #
    #   process_features_...
    #

    def test_process_features(self, processor, tmpdir):
        ds = resources.create_dataset()

        in_feat_path = os.path.join(tmpdir.strpath, 'in_feats')
        out_feat_path = os.path.join(tmpdir.strpath, 'out_feats')

        in_feats = containers.FeatureContainer(in_feat_path)
        utt_feats = np.arange(30).reshape(5, 6)

        with in_feats:
            in_feats.sampling_rate = 16000
            in_feats.frame_size = 400
            in_feats.hop_size = 160

            for utt_idx in ds.utterances.keys():
                in_feats.set(utt_idx, utt_feats)

        processor.process_features(ds, in_feats, out_feat_path)

        out_feats = containers.FeatureContainer(out_feat_path)

        with out_feats:
            assert len(out_feats.keys()) == 5

            assert np.array_equal(out_feats.get('utt-1', mem_map=False), utt_feats)
            assert np.array_equal(out_feats.get('utt-2', mem_map=False), utt_feats)
            assert np.array_equal(out_feats.get('utt-3', mem_map=False), utt_feats)
            assert np.array_equal(out_feats.get('utt-4', mem_map=False), utt_feats)
            assert np.array_equal(out_feats.get('utt-5', mem_map=False), utt_feats)

    def test_process_features_online(self, processor, tmpdir):
        ds = resources.create_dataset()

        in_feat_path = os.path.join(tmpdir.strpath, 'in_feats')
        out_feat_path = os.path.join(tmpdir.strpath, 'out_feats')

        in_feats = containers.FeatureContainer(in_feat_path)
        utt_feats = np.arange(30).reshape(5, 6)

        with in_feats:
            in_feats.sampling_rate = 16000
            in_feats.frame_size = 400
            in_feats.hop_size = 160

            for utt_idx in ds.utterances.keys():
                in_feats.set(utt_idx, utt_feats)

        processor.process_features_online(ds, in_feats, out_feat_path)

        out_feats = containers.FeatureContainer(out_feat_path)

        with out_feats:
            assert len(out_feats.keys()) == 5

            assert np.array_equal(out_feats.get('utt-1', mem_map=False), utt_feats)
            assert np.array_equal(out_feats.get('utt-2', mem_map=False), utt_feats)
            assert np.array_equal(out_feats.get('utt-3', mem_map=False), utt_feats)
            assert np.array_equal(out_feats.get('utt-4', mem_map=False), utt_feats)
            assert np.array_equal(out_feats.get('utt-5', mem_map=False), utt_feats)

    def test_process_features_online_with_given_chunk_size(self, processor, tmpdir):
        ds = resources.create_dataset()

        in_feat_path = os.path.join(tmpdir.strpath, 'in_feats')
        out_feat_path = os.path.join(tmpdir.strpath, 'out_feats')

        in_feats = containers.FeatureContainer(in_feat_path)
        utt_feats = np.arange(90).reshape(15, 6)

        with in_feats:
            in_feats.sampling_rate = 16000
            in_feats.frame_size = 400
            in_feats.hop_size = 160

            for utt_idx in ds.utterances.keys():
                in_feats.set(utt_idx, utt_feats)

        processor.process_features_online(ds, in_feats, out_feat_path, chunk_size=4)

        out_feats = containers.FeatureContainer(out_feat_path)

        assert len(processor.called_with_data) == 4 * 5
        assert processor.called_with_data[0].shape == (4, 6)
        assert processor.called_with_data[3].shape == (3, 6)

        with out_feats:
            assert len(out_feats.keys()) == 5

            assert np.array_equal(out_feats.get('utt-1', mem_map=False), utt_feats)
            assert np.array_equal(out_feats.get('utt-2', mem_map=False), utt_feats)
            assert np.array_equal(out_feats.get('utt-3', mem_map=False), utt_feats)
            assert np.array_equal(out_feats.get('utt-4', mem_map=False), utt_feats)
            assert np.array_equal(out_feats.get('utt-5', mem_map=False), utt_feats)

    def test_process_features_online_ignores_none(self, processor, tmpdir):
        ds = resources.create_dataset()

        in_feat_path = os.path.join(tmpdir.strpath, 'in_feats')
        out_feat_path = os.path.join(tmpdir.strpath, 'out_feats')

        in_feats = containers.FeatureContainer(in_feat_path)
        utt_feats = np.arange(90).reshape(15, 6)

        with in_feats:
            in_feats.sampling_rate = 16000
            in_feats.frame_size = 400
            in_feats.hop_size = 160

            for utt_idx in ds.utterances.keys():
                in_feats.set(utt_idx, utt_feats)

        def return_none(*args, **kwargs):
            return None

        processor.process_frames = return_none
        processor.process_features_online(ds, in_feats, out_feat_path, chunk_size=4)

        assert True

    def test_process_features_with_frame_hop_size_change_stores_correct(self, processor, tmpdir):
        ds = resources.create_dataset()

        in_feat_path = os.path.join(tmpdir.strpath, 'in_feats')
        out_feat_path = os.path.join(tmpdir.strpath, 'out_feats')

        in_feats = containers.FeatureContainer(in_feat_path)
        utt_feats = np.arange(30).reshape(5, 6)

        with in_feats:
            in_feats.sampling_rate = 16000
            in_feats.frame_size = 400
            in_feats.hop_size = 160

            for utt_idx in ds.utterances.keys():
                in_feats.set(utt_idx, utt_feats)

        processor.mock_frame_size_scale = 2.0
        processor.mock_hop_size_scale = 2.0
        processor.process_features(ds, in_feats, out_feat_path)

        out_feats = containers.FeatureContainer(out_feat_path)

        with out_feats:
            assert out_feats.frame_size == 800
            assert out_feats.hop_size == 320

    def test_process_features_online_with_frame_hop_size_change_stores_correct(self, processor, tmpdir):
        ds = resources.create_dataset()

        in_feat_path = os.path.join(tmpdir.strpath, 'in_feats')
        out_feat_path = os.path.join(tmpdir.strpath, 'out_feats')

        in_feats = containers.FeatureContainer(in_feat_path)
        utt_feats = np.arange(30).reshape(5, 6)

        with in_feats:
            in_feats.sampling_rate = 16000
            in_feats.frame_size = 400
            in_feats.hop_size = 160

            for utt_idx in ds.utterances.keys():
                in_feats.set(utt_idx, utt_feats)

        processor.mock_frame_size_scale = 2.0
        processor.mock_hop_size_scale = 2.0
        processor.process_features_online(ds, in_feats, out_feat_path)

        out_feats = containers.FeatureContainer(out_feat_path)

        with out_feats:
            assert out_feats.frame_size == 800
            assert out_feats.hop_size == 320
